import logging

from sfdata_stream_parser.checks import type_check
from sfdata_stream_parser import events
from sfdata_stream_parser.filters.generic import streamfilter, pass_event

from liiatools.datasets.s903.lds_ssda903_clean.converters import to_category, to_integer

from liiatools.datasets.shared_functions.converters import to_date
from liiatools.datasets.shared_functions.common import check_postcode

log = logging.getLogger(__name__)


@streamfilter(
    check=type_check(events.Cell), fail_function=pass_event, error_function=pass_event
)
def clean_dates(event):
    """
    Convert all values that should be dates to dates based on the config.yaml file

    :param event: A filtered list of event objects of type Cell
    :return: An updated list of event objects
    """
    date = event.config_dict["date"]
    try:
        text = to_date(event.cell, date)
        return event.from_event(event, cell=text, error="0")
    except (AttributeError, TypeError, ValueError):
        return event.from_event(event, cell="", error="1")


@streamfilter(
    check=type_check(events.Cell), fail_function=pass_event, error_function=pass_event
)
def clean_categories(event):
    """
    Convert all values that should be categories to categories based on the config.yaml file

    :param event: A filtered list of event objects of type Cell
    :return: An updated list of event objects
    """
    category = event.config_dict["category"]
    try:
        text = to_category(event.cell, category)
        if text != "error":
            return event.from_event(event, cell=text, error="0")
        else:
            return event.from_event(event, cell="", error="1")
    except (AttributeError, TypeError, ValueError):
        return event.from_event(event, cell="", error="1")


@streamfilter(
    check=type_check(events.Cell), fail_function=pass_event, error_function=pass_event
)
def clean_integers(event):
    """
    Convert all values that should be integers into integers based on the config.yaml file

    :param event: A filtered list of event objects of type Cell
    :return: An updated list of event objects
    """
    numeric = event.config_dict["numeric"]
    try:
        text = to_integer(event.cell, numeric)
        return event.from_event(event, cell=text, error="0")
    except (AttributeError, TypeError, ValueError):
        return event.from_event(event, cell="", error="1")


@streamfilter(
    check=lambda x: x.get("header") in ["HOME_POST", "PL_POST"],
    fail_function=pass_event,
)
def clean_postcodes(event):
    """
    Check that all values that should be postcodes are postcodes

    :param event: A filtered list of event objects of type Cell
    :return: An updated list of event objects
    """
    error = "0"
    text = ""
    try:
        text = check_postcode(event.cell)
    except (AttributeError, TypeError, ValueError):
        error = "1"
    return event.from_event(event, cell=text, error=error)


def clean(stream):
    """
    Compile the cleaning functions
    """
    stream = clean_dates(stream)
    stream = clean_categories(stream)
    stream = clean_integers(stream)
    stream = clean_postcodes(stream)
    return stream



import logging
from xmlschema import XMLSchemaValidatorError
import re

from sfdata_stream_parser.checks import type_check
from sfdata_stream_parser import events
from sfdata_stream_parser.filters.generic import streamfilter, pass_event
from sfdata_stream_parser.collectors import collector, block_check

log = logging.getLogger(__name__)


def _get_validation_error(
    event, schema, node, LAchildID_error, field_error
) -> XMLSchemaValidatorError:
    """
    Validate an event

    :param event: A filtered list of event objects
    :param schema: The xml schema attached to a given event
    :param node: The node attached to a given event
    :param LAchildID_error: An empty list to save child ID errors
    :param field_error: An empty list to save field errors
    :return: None if valid, event and error information if XMLSchemaValidatorError, event if AttributeError
    """
    try:
        schema.validate(node)
        return None
    except XMLSchemaValidatorError as e:
        regline = re.compile(
            r"(?=\(line.*?(\w+))", re.MULTILINE
        )  # Search for the number after "line" in error
        missing_field_line = regline.search(str(e)).group(1)
        if "Unexpected" and "LAchildID" in e.reason:
            LAchildID_error.append(
                f"LAchildID is missing from the node starting on line: {missing_field_line}, other errors associated "
                f"to this LAchildID will not be shown because of this"
            )
            return event
        if " expected" in e.reason:
            regexp = re.compile(
                r"(?=\sTag.*?(\w+))"
            )  # Search for the first word after "Tag"
            missing_field = regexp.search(e.reason).group(1)
            field_error.append(
                f"Missing required field: '{missing_field}' which occurs in the node starting "
                f"on line: {missing_field_line}"
            )
            return event
        else:
            if "failed validating ''" in e.message:
                return event.from_event(event, reason="blank")
            else:
                return event
    except AttributeError:  # Return event information, so it can be written to a log for the Local Authority
        return event


@streamfilter(check=type_check(events.StartElement), fail_function=pass_event)
def validate_elements(event, LAchildID_error, field_error):
    """
    Validates each element, and if not valid, sets the properties:

    :param event: A filtered list of event objects
    :param LAchildID_error: An empty list to save child ID errors
    :param field_error: An empty list to save field errors

    * valid - (always False)
    * validation_message - a descriptive validation message
    """
    validation_error = _get_validation_error(
        event, event.schema, event.node, LAchildID_error, field_error
    )
    if validation_error is None:
        return event

    if hasattr(validation_error, "reason"):
        message = validation_error.reason
        return events.StartElement.from_event(
            event, valid=False, validation_message=message
        )
    else:
        return events.StartElement.from_event(event, valid=False)


@collector(check=block_check(events.StartElement), receive_stream=True)
def remove_invalid(stream, tag_list):
    """
    Filters out events with the given tag name if they are not valid

    :param stream: A filtered list of event objects
    :param tag_list: A list of node tags
    :return: An updated list of event objects
    """
    stream = list(stream)
    first = stream[0]
    last = stream[-1]
    stream = stream[1:-1]

    if first.tag in tag_list and not getattr(first, "valid", True):
        yield from []
    else:
        yield first

        if len(stream) > 0:
            yield from remove_invalid(stream, tag_list=tag_list)

        yield last



import logging
from pathlib import Path
import pandas as pd
from datetime import datetime
import numpy as np

log = logging.getLogger(__name__)


def read_file(input, dates):
    """
    Reads the csv file as a pandas DataFrame
    """
    flatfile = pd.read_csv(input, parse_dates=dates, dayfirst=True)
    return flatfile


def _merge_dfs(flatfile, old_df, la_name):
    """
    Deletes existing data for new LA from pan file
    Merges new LA data to pan file
    """
    old_df = old_df.drop(old_df[old_df["LA"] == la_name].index)
    flatfile = pd.concat([flatfile, old_df], axis=0, ignore_index=True)
    return flatfile


def merge_agg_files(flat_output, dates, la_name, flatfile):
    """
    Checks if pan file exists
    Passes old and new file to function to be merged
    """
    output_file = Path(flat_output, f"pan_London_CIN_flatfile.csv")
    if output_file.is_file():
        old_df = pd.read_csv(output_file, parse_dates=dates, dayfirst=True)
        flatfile = _merge_dfs(flatfile, old_df, la_name)
    return flatfile


def export_flatfile(flat_output, flatfile):
    output_path = Path(flat_output, f"pan_London_CIN_flatfile.csv")
    flatfile.to_csv(output_path, index=False)


def filter_flatfile(flatfile, filter):
    """
    Filters rows to specified events
    Removes redundant columns that relate to other types of event
    """
    filtered_flatfile = flatfile[flatfile["Type"] == filter]
    filtered_flatfile = filtered_flatfile.dropna(axis=1, how="all")
    return filtered_flatfile


def split_factors(factors):
    """
    Creates a new set of columns from the flatfile with a column for each assessment factor
    Rows correspond the the rows of the flatfile and should have a value of 0 or 1 for each column
    """
    factor_cols = factors.Factors
    factor_cols = factor_cols.str.split(",", expand=True)
    factor_cols = factor_cols.stack()
    factor_cols = factor_cols.str.get_dummies()
    factor_cols = factor_cols.groupby(level=0).sum()
    assert factor_cols.isin([0, 1]).all(axis=None)
    factors = pd.concat([factors, factor_cols], axis=1)
    return factors


def export_factfile(analysis_output, factors):
    """
    Writes the factors output as a csv
    """
    output_path = Path(analysis_output, f"CIN_Census_factors.csv")
    factors.to_csv(output_path, index=False)


def referral_inputs(flatfile):
    """
    Creates three inputs for referral journeys analysis file
    """
    ref = filter_flatfile(flatfile, filter="CINreferralDate")
    s17 = filter_flatfile(flatfile, filter="AssessmentActualStartDate")
    s47 = filter_flatfile(flatfile, filter="S47ActualStartDate")
    return ref, s17, s47


def _time_between_date_series(later_date_series, earlier_date_series, years=0, days=0):
    days_series = later_date_series - earlier_date_series
    days_series = days_series.dt.days

    if days == 1:
        return days_series

    elif years == 1:
        years_series = (days_series / 365).apply(np.floor)
        years_series = years_series.astype('Int32')
        return years_series


def _filter_event_series(dataset, days_series, max_days):

    dataset = dataset[
        ((dataset[days_series] <= max_days) & (dataset[days_series] >= 0))
    ]
    return dataset


def merge_ref_s17(ref, s17, ref_assessment):
    """
    Merges ref and s17 views together, keeping only logically valid matches
    """
    # Merges referrals and assessments
    ref_s17 = ref.merge(
        s17[["LAchildID", "AssessmentActualStartDate"]], how="left", on="LAchildID"
    )

    # Calculates days between assessment and referral
    ref_s17["days_to_s17"] = _time_between_date_series(
        ref_s17["AssessmentActualStartDate"], ref_s17["CINreferralDate"], days=1
    )

    # Only assessments within config-specifed period following referral are valid
    ref_s17 = _filter_event_series(ref_s17, "days_to_s17", ref_assessment)

    # Reduces dataset to fields required for analysis
    ref_s17 = ref_s17[["Date", "LAchildID", "AssessmentActualStartDate", "days_to_s17"]]

    return ref_s17


def merge_ref_s47(ref, s47, ref_assessment):
    """
    Merges ref and s47 views together, keeping only logically valid matches
    """
    # Merges referrals and S47s
    ref_s47 = ref.merge(
        s47[["LAchildID", "S47ActualStartDate"]], how="left", on="LAchildID"
    )

    # Calculates days between S47 and referral
    ref_s47["days_to_s47"] = _time_between_date_series(
        ref_s47["S47ActualStartDate"], ref_s47["CINreferralDate"], days=1
    )

    # Only S47s within config-specifed period following referral are valid
    ref_s47 = _filter_event_series(ref_s47, "days_to_s47", ref_assessment)

    # Reduces dataset to fields required for analysis
    ref_s47 = ref_s47[["Date", "LAchildID", "S47ActualStartDate", "days_to_s47"]]

    return ref_s47


def ref_outcomes(ref, ref_s17, ref_s47):
    """
    Merges views together to give all outcomes of referrals in one place
    Outcomes column defaults to NFA unless there is a relevant S17 or S47 event to match
    Calculates age of child at referral
    """
    # Merge databases together
    ref_outs = ref.merge(ref_s17, on=["Date", "LAchildID"], how="left")
    ref_outs = ref_outs.merge(ref_s47, on=["Date", "LAchildID"], how="left")

    # Set default outcome to "NFA"
    ref_outs["referral_outcome"] = "NFA"

    # Set outcome to "S17" when there is a relevant assessment
    ref_outs.loc[
        ref_outs["AssessmentActualStartDate"].notnull(), "referral_outcome"
    ] = "S17"

    # Set outcome to "S47" when there is a relevant S47
    ref_outs.loc[ref_outs["S47ActualStartDate"].notnull(), "referral_outcome"] = "S47"

    # Set outcome to "Both S17 & S47" when there are both
    ref_outs.loc[
        (
            (ref_outs["AssessmentActualStartDate"].notnull())
            & (ref_outs["S47ActualStartDate"].notnull())
        ),
        "referral_outcome",
    ] = "Both S17 & S47"

    # Calculate age of child at referral
    ref_outs["Age at referral"] = _time_between_date_series(
        ref_outs["CINreferralDate"], ref_outs["PersonBirthDate"], years=1
    )

    return ref_outs


def export_reffile(analysis_output, ref_outs):
    """
    Writes the referral journeys output as a csv
    """
    output_path = Path(analysis_output, f"CIN_Census_referrals.csv")
    ref_outs.to_csv(output_path, index=False)


def journey_inputs(flatfile):
    """
    Creates the input for the journey analysis file
    """
    # Create inputs from flatfile and merge them
    s47_j = filter_flatfile(flatfile, "S47ActualStartDate")
    cpp = filter_flatfile(flatfile, "CPPstartDate")
    return s47_j, cpp


def journey_merge(s47_j, cpp, icpc_cpp_days, s47_cpp_days):
    """
    Merges inputs to produce outcomes file
    """
    s47_cpp = s47_j.merge(
        cpp[["LAchildID", "CPPstartDate"]], how="left", on="LAchildID"
    )

    # Calculate days from ICPC to CPP start
    s47_cpp["icpc_to_cpp"] = _time_between_date_series(
        s47_cpp["CPPstartDate"], s47_cpp["DateOfInitialCPC"], days=1
    )

    # Calculate days from S47 to CPP start
    s47_cpp["s47_to_cpp"] = _time_between_date_series(
        s47_cpp["CPPstartDate"], s47_cpp["S47ActualStartDate"], days=1
    )

    # Only keep logically consistent events (as defined in config variables)
    s47_cpp = s47_cpp[
        ((s47_cpp["icpc_to_cpp"] >= 0) & (s47_cpp["icpc_to_cpp"] <= icpc_cpp_days))
        | ((s47_cpp["s47_to_cpp"] >= 0) & (s47_cpp["s47_to_cpp"] <= s47_cpp_days))
    ]

    # Merge events back to S47_j view
    s47_outs = s47_j.merge(
        s47_cpp[["Date", "LAchildID", "CPPstartDate", "icpc_to_cpp", "s47_to_cpp"]],
        how="left",
        on=["Date", "LAchildID"],
    )

    return s47_outs


def s47_paths(s47_outs, s47_day_limit, icpc_day_limit):
    """
    Creates an output that can generate a Sankey diagram of outcomes from S47 events
    """
    # Dates used to define window for S47 events where outcome may not be known because CIN Census is too recent
    for y in s47_outs["YEAR"]:
        s47_outs["cin_census_close"] = datetime(int(y), 3, 31)
    s47_outs["s47_max_date"] = s47_outs["cin_census_close"] - pd.Timedelta(
        s47_day_limit
    )
    s47_outs["icpc_max_date"] = s47_outs["cin_census_close"] - pd.Timedelta(
        icpc_day_limit
    )

    # Setting the Sankey diagram source for S47 events
    step1 = s47_outs.copy()
    step1["Source"] = "S47 strategy discussion"

    # Setting the Sankey diagram destination for S47 events
    step1["Destination"] = np.nan

    step1.loc[step1["DateOfInitialCPC"].notnull(), "Destination"] = "ICPC"

    step1.loc[
        step1["DateOfInitialCPC"].isnull() & step1["CPPstartDate"].notnull(),
        "Destination",
    ] = "CPP start"

    step1.loc[
        (
            (step1["Destination"].isnull())
            & (step1["S47ActualStartDate"] >= step1["s47_max_date"])
        ),
        "Destination",
    ] = "TBD - S47 too recent"

    step1.loc[step1["Destination"].isnull(), "Destination"] = "No ICPC or CPP"

    # Setting the Sankey diagram source for ICPC events
    step2 = step1[step1["Destination"] == "ICPC"]
    step2["Source"] = "ICPC"

    # Setting the Sankey diagram destination for ICPC events
    step2["Destination"] = np.nan

    step2.loc[step2["CPPstartDate"].notnull(), "Destination"] = "CPP start"

    step2.loc[
        (
            (step2["Destination"].isnull())
            & (step2["DateOfInitialCPC"] >= step2["icpc_max_date"])
        ),
        "Destination",
    ] = "TBD - ICPC too recent"

    step2.loc[step2["Destination"].isnull(), "Destination"] = "No CPP"

    # Merge the steps together
    s47_journey = pd.concat([step1, step2])

    # Calculate age of child at S47
    s47_journey["Age at S47"] = _time_between_date_series(
        s47_journey["S47ActualStartDate"], s47_journey["PersonBirthDate"], years=1
    )

    return s47_journey


def export_journeyfile(analysis_output, s47_journey):
    """
    Writes the S47 journeys output as a csv
    """
    output_path = Path(analysis_output, f"CIN_Census_S47_journey.csv")
    s47_journey.to_csv(output_path, index=False)



"""
   Spreadsheet generation with columns ordered by: YearCensus, LEAName, Gender, Ethnicity_Compact. 
   And with the sum columns: FTESum. 
   And the employee count: SWENo_Count
"""
import os
import pandas as pd
import liiatools.datasets.social_work_workforce.SWFtools.util.work_path as work_path


def pivotGen():
    # ===== Read file ===== #
    file = "merged_modified.csv"
    path = work_path.flatfile_folder
    requestPath = work_path.request
    pathFile = os.path.join(path, file)
    df = pd.read_csv(pathFile)

    pivotTable = df.groupby(
        ["YearCensus", "LEAName", "Gender", "Ethnicity_Compact"]
    ).agg(FTESum=("FTE", "sum"), SWENo_Count=("SWENo", "count"))

    # ===== Save and export file ===== #
    fileOutN = "pivotTable.xlsx"
    requestPath = work_path.request
    fileOut = os.path.join(requestPath, fileOutN)
    pivotTable.to_excel(fileOut, merge_cells=False)



import logging
import click as click
import click_log

from liiatools.datasets.social_work_workforce import csww_main_functions


logger = logging.getLogger()
click_log.basic_config(logger)


@click.group()
def csww():
    """
    Functions for creating CSWW Census sample file generator
    """
    pass


@csww.command()
@click.option(
    "--o",
    "output",
    required=True,
    type=str,
    help="A string specifying the output file location, including the file name and suffix",
)
@click_log.simple_verbosity_option(logger)
def generate_sample(output: str):
    """
    Export a sample file for testing

    :param output: string containing the desired location and name of sample file
    :return: .xml sample file in desired location
    """
    output = csww_main_functions.generate_sample(output)
    return output



from functools import cached_property
from pathlib import Path

import xmlschema

from liiatools.spec import cin_census as cin_asset_dir


class Schema:
    def __init__(self, year):
        self.__year = year

    @cached_property
    def schema(self) -> xmlschema.XMLSchema:
        return xmlschema.XMLSchema(
            Path(cin_asset_dir.__file__).parent / f"CIN_schema_{self.__year}.xsd"
        )



from collections import Counter
from datetime import datetime
import logging
import os

from sfdata_stream_parser import events
from sfdata_stream_parser.filters.generic import streamfilter, pass_event
from sfdata_stream_parser.checks import type_check

log = logging.getLogger(__name__)


class ErrorTable(events.ParseEvent):
    pass


def create_error_table(stream):
    """
    Create an ErrorTable event containing the sheet name for that table

    :param stream: A filtered list of event objects
    :return: An updated list of event objects with error counts
    """
    sheet_name = None
    for event in stream:
        if isinstance(event, events.StartTable):
            sheet_name = getattr(event, "sheet_name", None)
        elif isinstance(event, events.EndTable):
            yield ErrorTable.from_event(
                event,
                sheet_name=sheet_name,
            )
        yield event


@streamfilter(
    check=type_check(events.Cell), fail_function=pass_event, error_function=pass_event
)
def blank_error_check(event):
    """
    Check all the values against the config to see if they are allowed to be blank
    if they are blank but should not be, record this as event.blank_error = 1

    :param event: A filtered list of event objects of type Cell
    :return: An updated list of event objects
    """
    try:
        allowed_blank = event.other_config["canbeblank"]
        formatting_error = getattr(event, "formatting_error", "0")
        if (
            not allowed_blank
            and not event.value
            and event.value != 0
            and formatting_error != "1"
            and getattr(event, "blank_row", 0) != "1"
        ):
            return event.from_event(event, blank_error="1")
        else:
            return event
    except KeyError:  # Raised in case there is no config item for the given cell
        pass


def create_error_list(stream, error_name):
    """
    Create a list of the column headers for cells with errors

    :param stream: A filtered list of event objects
    :param error_name: A string containing the error_name to inherit e.g. "blank_error"
    :return: An updated list of event objects
    """
    error_list = None
    for event in stream:
        if isinstance(event, events.StartTable):
            error_list = []
        elif isinstance(event, ErrorTable):
            yield ErrorTable.from_event(event, **{f"{error_name}_list": error_list})
            error_list = None
        elif error_list is not None and isinstance(event, events.Cell):
            if getattr(event, error_name, None) == "1":
                error_list.append(event.column_header)
        yield event


def _duplicate_columns(columns_list):
    """
    Return a list of duplicate items within a given list

    :param columns_list: A list of values
    :return: A list of duplicated values
    """
    unique_columns = set()
    duplicate_columns = []
    for column in columns_list:
        if column not in unique_columns:
            unique_columns.add(column)
        else:
            duplicate_columns.append(column)
    return duplicate_columns


@streamfilter(
    check=type_check(events.StartTable),
    fail_function=pass_event,
    error_function=pass_event,
)
def duplicate_column_check(event):
    """
    Create a duplicate_columns_error object for any StartTable that has correctly matched columns but some duplicates

    :param event: A filtered list of event objects of type StartTable
    :return: An updated list of event objects
    """
    try:
        column_headers = event.matched_column_headers
        if len(set(column_headers)) == len(column_headers):
            return event
        else:
            duplicate_columns = _duplicate_columns(column_headers)
            duplicate_columns = str(duplicate_columns)[
                1:-1
            ]  # "Remove [ and ] from string
            return event.from_event(
                event,
                duplicate_columns=f"Sheet with title {event.sheet_name} contained "
                f"the following duplicate column(s): "
                f"{duplicate_columns}",
            )
    except KeyError:  # Raised in case there is no matched_column_headers
        pass
    return event


def inherit_error(stream, error_name):
    """
    Add the error_name_error value to the ErrorTable so these errors can be written to the log.txt file

    :param stream: A filtered list of event objects
    :param error_name: A string containing the error_name to inherit e.g. "duplicate_columns"
    :return: An updated list of event objects
    """
    error = []
    for event in stream:
        try:
            if isinstance(event, events.StartTable):
                error = getattr(event, error_name, [])
            elif isinstance(event, events.EndTable):
                error = []
            elif isinstance(event, ErrorTable):
                yield ErrorTable.from_event(event, **{f"{error_name}_error": error})
            yield event
        except AttributeError:
            yield event


@streamfilter(
    check=type_check(events.StartTable),
    fail_function=pass_event,
    error_function=pass_event,
)
def create_file_match_error(event):
    """
    Add a match_error to StartTables that do not have an event.sheet_name so these errors can be written to the log.txt
    file. If there is no event.sheet_name for a given StartTable that means its headers did not match any of those
    in the config file

    :param event: A filtered list of event objects
    :return: An updated list of event objects
    """
    try:
        if event.sheet_name:
            return event
    except AttributeError:
        return event.from_event(
            event,
            match_error=f"Failed to find a set of matching columns headers for sheet titled "
            f"'{event.name}' which contains column headers "
            f"{event.column_headers}",
        )
    return event


def _missing_sheet_match(sheet_names, expected_sheet_names):
    """
    Return a list of sheet_names missing from the expected_sheet_names

    :param sheet_names: A list of sheet names
    :param expected_sheet_names: A list of expected sheet names
    :return: A list of sheet names missing from the expected sheet names
    """
    return list(set(expected_sheet_names).difference(sheet_names))


def create_missing_sheet_error(stream):
    """
    Checks for any missing sheet names against a list of expected sheet names

    :param stream: A filtered list of event objects
    :return: An updated list of event objects and a list of missing sheets
    """
    sheet_names = []
    expected_sheet_names = [
        "List 1",
        "List 2",
        "List 3",
        "List 4",
        "List 5",
        "List 6",
        "List 7",
        "List 8",
        "List 9",
        "List 10",
        "List 11",
    ]
    for event in stream:
        if isinstance(event, events.StartTable):
            sheet_name = getattr(event, "sheet_name", None)
            sheet_names.append(sheet_name)
        if isinstance(event, events.EndContainer):
            missing_sheet_error = _missing_sheet_match(
                sheet_names, expected_sheet_names
            )
            if missing_sheet_error:
                missing_sheet_error = str(missing_sheet_error)[
                    1:-1
                ]  # Remove brackets [] from list
                yield event.from_event(
                    event,
                    missing_sheet_error=f"The following sheets are missing: "
                    f"{missing_sheet_error} so no output has been "
                    f"created",
                )
            else:
                yield event
        else:
            yield event


def save_errors_la(stream, la_log_dir):
    """
    Count the error events and save them as a text file in the Local Authority Logs directory
    only save the error events if there is at least one error in said event

    :param stream: A filtered list of event objects
    :param la_log_dir: Location to save the gathered error logs
    :return: An updated list of event objects
    """
    start_time = f"{datetime.now():%Y-%m-%dT%H%M%SZ}"
    for event in stream:
        try:
            if isinstance(event, ErrorTable) and (
                event.formatting_error_list is not None
                and event.blank_error_list is not None
                and event.sheet_name is not None
                and event.extra_columns_error is not None
                and event.duplicate_columns_error is not None
            ):
                if (
                    event.formatting_error_list
                    or event.blank_error_list
                    or event.extra_columns_error
                    or event.duplicate_columns_error
                ):
                    with open(
                        f"{os.path.join(la_log_dir, event.filename)}_error_log_{start_time}.txt",
                        "a",
                    ) as f:
                        f.write("\n")
                        f.write(event.sheet_name)
                        f.write("\n")
                        if event.formatting_error_list:
                            f.write(
                                "Number of cells that have been made blank "
                                "because they could not be formatted correctly"
                            )
                            f.write("\n")
                            counter_dict = Counter(event.formatting_error_list)
                            f.write(
                                str(counter_dict)[9:-2]
                            )  # Remove "Counter({" and "})" from string
                            f.write("\n")
                        if event.blank_error_list:
                            f.write(
                                "Number of blank cells that should have contained data"
                            )
                            f.write("\n")
                            blank_counter_dict = Counter(event.blank_error_list)
                            f.write(
                                str(blank_counter_dict)[9:-2]
                            )  # Remove "Counter({" and "})" from string
                            f.write("\n")
                        if event.extra_columns_error:
                            extra_columns_error_no_none = list(
                                filter(None, event.extra_columns_error)
                            )
                            f.write(
                                f"Headers of unexpected columns that have been removed or reformatted: "
                                f"{extra_columns_error_no_none}"
                            )
                            f.write("\n")
                        if event.duplicate_columns_error:
                            f.write(event.duplicate_columns_error)
                            f.write("\n")
        except AttributeError:
            pass

        try:
            if isinstance(event, events.StartTable) and event.match_error is not None:
                with open(
                    f"{os.path.join(la_log_dir, event.filename)}_error_log_{start_time}.txt",
                    "a",
                ) as f:
                    f.write("\n")
                    f.write(event.match_error)
                    f.write("\n")
        except AttributeError:
            pass

        try:
            if (
                isinstance(event, events.EndContainer)
                and event.missing_sheet_error is not None
            ):
                with open(
                    f"{os.path.join(la_log_dir, event.filename)}_error_log_{start_time}.txt",
                    "a",
                ) as f:
                    f.write("\n")
                    f.write(event.missing_sheet_error)
                    f.write("\n")
        except AttributeError:
            pass

        yield event


def log_errors(stream):
    """
    Compile the log error functions

    :param stream: A filtered list of event objects
    :return: An updated list of event objects
    """
    stream = create_error_table(stream)
    stream = blank_error_check(stream)
    stream = create_error_list(stream, error_name="formatting_error")
    stream = create_error_list(stream, error_name="blank_error")
    stream = inherit_error(stream, error_name="extra_columns")
    stream = duplicate_column_check(stream)
    stream = inherit_error(stream, error_name="duplicate_columns")
    stream = create_file_match_error(stream)
    stream = create_missing_sheet_error(stream)
    return stream



from datetime import datetime, date
import re


def to_date(datevalue, dateformat="%d/%m/%Y"):
    """
    Convert a string to a date based on the dateformat %d/%m/%Y and convert a datetime to a date
    :param datevalue: A value to test and make sure it's a datetime object
    :param dateformat: A format for the date to be read correctly, default to %d/%m/%Y
    :return: Either the specified date, converted to a datetime, or an empty string
    """
    if datevalue:
        if isinstance(datevalue, datetime):
            return datevalue.date()
        if isinstance(datevalue, date):
            return datevalue
        elif isinstance(datevalue, str):
            return datetime.strptime(datevalue, dateformat).date()
    else:
        return ""


def to_short_postcode(postcode):
    """
    Remove whitespace from the beginning and end of postcodes and the last two digits for anonymity
    return blank if not in the right format
    :param postcode: A string with a UK-style post code
    :return: a shortened post code with the area, district, and sector. The units is removed
    """
    try:
        match = re.search(
            r"([A-Z]+\d\d?[A-Z]?)\s*(\d+)([A-Z]{2})", postcode, re.IGNORECASE
        )
        postcode = match.group(1) + " " + match.group(2)
    except (AttributeError, TypeError, ValueError):
        postcode = ""
    return postcode


def to_month_only_dob(dob):
    """
    Convert dates of birth into month and year of birth starting from 1st of each month for anonymity
    return blank if not in the right format
    :param dob: A date of birth datetime object
    :return: A date of birth datetime object with the month rounded to the first day
    """
    try:
        dob = dob.replace(day=1)
    except (AttributeError, TypeError, ValueError):
        dob = ""
    return dob



import logging
import pandas as pd
from pathlib import Path
import cchardet as chardet
from datetime import datetime

import pandas.errors


def file_encoding_audit(
    data_folder: Path,
) -> pd.DataFrame:
    """
    Function takes in a folder path object, it then uses the cchardet library to fast detect the file encoding types

    :param data_folder: Path object that is a folder containing files to be processed
    :type data_folder: Path
    :return: A Dataframe of low confidence encoded files
    :rtype: pd.DataFrame
    """

    # TODO - Check csv encoding type of file, save to utf-8
    # TODO - Check xml encoding type of file
    # Save as an acceptable format
    result_out = []

    for cdf in data_folder.glob("**/*"):
        if cdf.is_file() and "log" not in cdf.root:
            with open(cdf, "rb") as f:
                msg = f.read()
                result = chardet.detect(msg)
                out = f"{cdf.parts[-3]}, {cdf.stem}, {result}"
                # this is messy.
                outt = (
                    out.replace("}", "")
                    .replace("{", "")
                    .replace("confidence", "")
                    .replace("encoding", "")
                    .replace("'':", "")
                )
                result_out.append(outt)

    # Save the outputs of the list generated by running cchardet on the file list,
    # the result is then appended into a dataframe and filtered to return a list of files that
    # have low confidence as to their encoding types.
    encoding_series = pd.Series(result_out)

    encoding_df = pd.DataFrame(encoding_series, columns=["file_name"])

    # Split out dataframe
    encoding_df[
        ["local_authority", "file_name", "encoding", "confidence"]
    ] = encoding_df.file_name.str.split(",", expand=True)

    # Filter out log files and drop high confidence files types
    encoded_df = encoding_df[
        ~encoding_df["file_name"].str.contains("Logs")
        & ~(encoding_df["confidence"].str.contains("1.0"))
    ]

    encoded_df.to_csv("encoding_audit.csv", encoding="utf-8")
    return encoded_df


def check_blank_file(input: str, la_log_dir: str):
    """
    Check csv file is not empty

    :param input: Path to file that needs to be checked
    :param la_log_dir: Location to save the error log
    :return: If csv is empty stop process and write log to local authority, else continue
    """
    start_time = f"{datetime.now():%d-%m-%Y %Hh-%Mm-%Ss}"
    input = Path(input)
    filename = input.resolve().stem
    extension = Path(input).suffix
    try:
        pd.read_csv(input)
        pass
    except pandas.errors.EmptyDataError:
        with open(
            f"{Path(la_log_dir, filename)}_error_log_{start_time}.txt",
            "a",
        ) as f:
            f.write(f"File: '{filename}{extension}' was found to be completely empty")
        return "empty"


def drop_empty_rows(infile: str, outfile: str):
    """
    csv drop empty rows at top of file, save output

    :param infile: Path to file that needs to be cleaned
    :type infile: String
    :param outfile: Path where cleaned file will be saved
    :type outfile: String
    """
    infile = Path(infile)
    data = pd.read_csv(infile, skip_blank_lines=True)
    logging.info(f"removing blank rows in {infile.stem}")
    data = data.dropna(how="all")
    data.to_csv(outfile, header=True, encoding="utf-8", index=False)
    return data


def delete_unrequired_files(input: str, drop_file_list: list, la_log_dir: str):
    """
    Deletes files if they match a substring name from list.

    :param input: should specify the input file location, including file name and suffix,
    and be usable by a Path function
    :type input: File
    :param drop_file_list: List of file names that will be removed
    :type drop_file_list: List
    :param la_log_dir: Path to the local authority's log folder
    :type la_log_dir: Path
    """
    input = Path(input)
    for dfl in drop_file_list:
        if dfl in input.stem:
            # logged in datapipe logs
            logging.info(f"{input.stem} removing file from processing not required.")
            save_unrequired_file_error(input, la_log_dir)
            input.unlink()
            raise Exception(
                f"{input.stem} has been deleted because it does not math the list of accepted files"
            )


def save_unrequired_file_error(input: Path, la_log_dir: str):
    """
    Saves unrequired file errors to a text file in the LA log directory

    :param input: The input file location, including file name and suffix, and be usable by a Path function
    :param la_log_dir: Path to the local authority's log folder
    :return: Text file containing the error information
    """
    filename = input.resolve().stem
    start_time = f"{datetime.now():%d-%m-%Y %Hh-%Mm-%Ss}"
    with open(
        f"{Path(la_log_dir, filename)}_error_log_{start_time}.txt",
        "a",
    ) as f:
        f.write(
            f"'{filename}' has been deleted because it does not match the list of accepted files"
        )



import random
import string
from datetime import date, datetime, timedelta
import pytz

from sfdata_stream_parser.events import StartElement, EndElement, TextNode

from liiatools.datasets.social_work_workforce.schema import Schema


def TextElement(tag: str, text):
    """
    Create a complete TextNode sandwiched between a StartElement and EndElement

    :param tag: XML tag
    :param text: text to be stored in the given XML tag, could be a string, integer, float etc.
    :return: StartElement and EndElement with given tags and TextNode with given text
    """
    yield StartElement(tag=tag)
    yield TextNode(text=str(text))
    yield EndElement(tag=tag)


def generate_sample_csww_file(**options):
    """
    Generate a sample children's social work workforce census file

    :return: stream of generators containing information required to create an XML file
    """
    yield StartElement(tag="Message")
    yield from generate_sample_header(**options)

    yield StartElement(tag="LALevelVacancies")
    yield from generate_la_level_vacancies()
    yield EndElement(tag="LALevelVacancies")

    for worker in range(random.randint(1, 50)):
        yield from generate_csww_worker()

    yield EndElement(tag="Message")


def generate_sample_header(**opts):
    """
    Generate information for the <Header> XML element

    :return: stream of generators containing <Header> information required to create an XML file
    """
    yield StartElement(tag="Header")
    yield StartElement(tag="CollectionDetails")
    yield from TextElement(tag="Collection", text="CSWW")
    yield from TextElement(tag="Year", text=opts.get("year", date.today().year))
    yield from TextElement(
        tag="ReferenceDate", text=opts.get("year", date.today().strftime("%Y-%m-%d"))
    )
    yield EndElement(tag="CollectionDetails")
    yield StartElement(tag="Source")
    yield from TextElement(tag="SourceLevel", text="L")
    if opts.get("lea"):
        yield from TextElement(tag="LEA", text=opts["lea"])
    else:
        yield from TextElement(tag="LEA", text=f"{random.randint(100, 999):03}")

    yield from TextElement(tag="SoftwareCode", text=__name__)
    yield from TextElement(
        tag="DateTime",
        text=datetime.now().astimezone(pytz.UTC).strftime("%Y-%m-%dT%H:%M:%SZ"),
    )

    yield EndElement(tag="Source")
    yield EndElement(tag="Header")


def generate_la_level_vacancies():
    """
    Generate information for the <LALevelVacancies> XML element

    :return: stream of generators containing <LALevelVacancies> information required to create an XML file
    """
    yield from TextElement(
        tag="NumberOfVacancies", text=round(random.uniform(0, 100), 2)
    )
    yield from TextElement(tag="NoAgencyFTE", text=round(random.uniform(0, 100), 2))
    yield from TextElement(tag="NoAgencyHeadcount", text=random.randint(0, 100))


def random_chance_0_1(prob_0=0.25, prob_1=0.75):
    """
    Create a random value of 0 or 1 using given probabilities of each occurring

    :param prob_0: probability of 0 occurring (default is 0.25)
    :param prob_1: probability of 1 occurring (default is 0.75)
    :return: integer of 0 or 1
    """
    assert prob_0 + prob_1 == 1
    return random.choices([0, 1], [prob_0, prob_1])[0]


def generate_csww_worker(configuration=None, **opts):
    """
    Generate information for the <CSWWWorker> XML element

    :param configuration: configuration .xsd file used to create elements matching the .xsd schema
    :return: stream of generators containing <CSWWWorker> information required to create an XML file
    """
    if not configuration:
        configuration = Configuration()

    if opts.get("seed"):
        random.seed(opts["seed"])

    dob = opts.get(
        "dob", date.today() - timedelta(days=random.uniform(365 * 18, 365 * 66))
    )

    agency_worker = configuration.random_agencyworker

    role_start_date = dob + timedelta(days=random.uniform(365 * 18, 365 * 50))
    if role_start_date > date.today():
        role_start_date = date.today()

    role_end_date_count = (
        random.choices([0, 1])[0] if role_start_date < date.today() else 0
    )
    role_end_date = role_start_date + timedelta(days=random.uniform(365 * 18, 365 * 50))
    if role_end_date > date.today():
        role_end_date = date.today()

    yield StartElement(tag="CSWWWorker")
    yield from TextElement(tag="AgencyWorker", text=agency_worker)
    yield from TextElement(
        tag="SWENo",
        text=f"{''.join(random.choices(string.ascii_letters, k=2))}"
        f"{random.randint(1000000000, 9999999999)}",
    )

    text = 0 if role_end_date_count == 1 else round(random.uniform(0, 1), 6)
    yield from TextElement(tag="FTE", text=text)

    if agency_worker == 0:
        yield from TextElement(tag="PersonBirthDate", text=dob.strftime("%Y-%m-%d"))
        yield from TextElement(tag="GenderCurrent", text=configuration.random_gender)
        yield from TextElement(tag="Ethnicity", text=configuration.random_ethnicity)
        yield from TextElement(tag="QualInst", text="Institution Name")
        yield from TextElement(tag="QualLevel", text=configuration.random_qual)
        yield from TextElement(tag="StepUpGrad", text=configuration.random_yesno)
        yield from TextElement(tag="OrgRole", text=configuration.random_role)
        yield from TextElement(
            tag="RoleStartDate", text=role_start_date.strftime("%Y-%m-%d")
        )
        yield from TextElement(tag="StartOrigin", text=configuration.random_origin)
        if role_end_date_count == 1:
            yield from TextElement(
                tag="RoleEndDate", text=role_end_date.strftime("%Y-%m-%d")
            )
            yield from TextElement(
                tag="LeaverDestination", text=configuration.random_leaver
            )
            yield from TextElement(tag="ReasonLeave", text=configuration.random_reason)
        yield from TextElement(tag="FTE30", text=round(random.uniform(0, 1), 6))
        yield from TextElement(tag="Cases30", text=random.randint(0, 100))
        yield from TextElement(
            tag="WorkingDaysLost", text=round(random.uniform(0, 100), 2)
        )
        yield from TextElement(
            tag="ContractWeeks", text=round(random.uniform(0, 500), 1)
        )
        yield from TextElement(tag="FrontlineGrad", text=configuration.random_yesno)

    else:
        if random_chance_0_1() == 1:
            yield from TextElement(tag="PersonBirthDate", text=dob.strftime("%Y-%m-%d"))
        if random_chance_0_1() == 1:
            yield from TextElement(
                tag="GenderCurrent", text=configuration.random_gender
            )
        if random_chance_0_1() == 1:
            yield from TextElement(tag="Ethnicity", text=configuration.random_ethnicity)
        if random_chance_0_1() == 1:
            yield from TextElement(tag="QualInst", text="Institution Name")
        if random_chance_0_1() == 1:
            yield from TextElement(tag="QualLevel", text=configuration.random_qual)
        if random_chance_0_1() == 1:
            yield from TextElement(tag="StepUpGrad", text=configuration.random_yesno)
        if random_chance_0_1() == 1:
            yield from TextElement(tag="OrgRole", text=configuration.random_role)
        if random_chance_0_1() == 1:
            yield from TextElement(
                tag="RoleStartDate", text=role_start_date.strftime("%Y-%m-%d")
            )
        if random_chance_0_1() == 1:
            yield from TextElement(tag="StartOrigin", text=configuration.random_origin)
        if role_end_date_count == 1:
            yield from TextElement(
                tag="RoleEndDate", text=role_end_date.strftime("%Y-%m-%d")
            )
            yield from TextElement(
                tag="LeaverDestination", text=configuration.random_leaver
            )
            yield from TextElement(tag="ReasonLeave", text=configuration.random_reason)
        if random_chance_0_1() == 1:
            yield from TextElement(tag="FTE30", text=round(random.uniform(0, 1), 6))
        if random_chance_0_1() == 1:
            yield from TextElement(tag="Cases30", text=random.randint(0, 100))
        if random_chance_0_1() == 1:
            yield from TextElement(
                tag="WorkingDaysLost", text=round(random.uniform(0, 100), 2)
            )
        if random_chance_0_1() == 1:
            yield from TextElement(
                tag="ContractWeeks", text=round(random.uniform(0, 500), 1)
            )
        if random_chance_0_1() == 1:
            yield from TextElement(tag="FrontlineGrad", text=configuration.random_yesno)

    if random_chance_0_1(0.8, 0.2) == 1:
        yield from TextElement(tag="Absat30Sept", text=configuration.random_yesno)
        yield from TextElement(tag="ReasonAbsence", text=configuration.random_absence)

    if random_chance_0_1(0.5, 0.5) == 1:
        yield from TextElement(tag="CFKSSstatus", text=configuration.random_cfkss)

    yield EndElement(tag="CSWWWorker")


class Configuration:
    def __init__(self, schema=None):
        if schema:
            self._schema = schema
        else:
            self._schema = Schema().schema

    def __getattr__(self, item: str):
        """
        Generate random value from a given element in the schema

        :param item: name of element we want to get a random value from
        :return: random value from a given element
        """
        if item.startswith("random_"):
            values = getattr(self, item[7:])
            if values is not None:
                return random.choice(values)
        else:
            value = self._schema.types[f"{item}type"]
            if value is not None:
                return value.enumeration
        raise AttributeError(f"{item} not found.")



from pathlib import Path
import pandas as pd
from datetime import date
import logging

log = logging.getLogger(__name__)


def split_file(input):
    """Reads xlsx file as dictionary of dataframes"""
    aa_dict = pd.read_excel(input, sheet_name=None, index_col=None, dtype=object)
    return aa_dict


def sort_dict(aa_dict, sort_order):
    """Sorts the imported dict by List to ensure consistency"""
    for k in aa_dict.keys():
        df = aa_dict[k]
        df = df[sort_order[k]]
        aa_dict[k] = df
    return aa_dict


def _merge_dfs(aa_dict, old_dict):
    for k in aa_dict.keys():
        new_df = aa_dict[k]
        old_df = old_dict[k]
        merged_df = pd.concat([new_df, old_df], axis=0, ignore_index=True)
        aa_dict[k] = merged_df
    return aa_dict


def merge_la_files(output, aa_dict):
    output_file = Path(output, f"AnnexA_merged.xlsx")
    if output_file.is_file():
        old_dict = pd.read_excel(
            output_file, sheet_name=None, index_col=None, dtype=object
        )
        _merge_dfs(aa_dict, old_dict)
    return aa_dict


def deduplicate(aa_dict, dedup):
    for k in aa_dict.keys():
        df = aa_dict[k]
        df = df.drop_duplicates(subset=dedup[k], keep="first")
        aa_dict[k] = df
    return aa_dict


def convert_datetimes(aa_dict, dates):
    for k in aa_dict.keys():
        df = aa_dict[k]
        for date_field in dates[k]:
            df[date_field] = pd.to_datetime(df[date_field], format="%d/%m/%Y")
        aa_dict[k] = df
    return aa_dict


def _remove_years(years):
    d = pd.to_datetime("today")
    try:
        return d.replace(year=d.year - years)
    except ValueError:
        return d + (date(d.year - years, 1, 1) - date(d.year, 1, 1))


def remove_old_data(aa_dict, index_date):
    for k in aa_dict.keys():
        index_date_key = index_date[k]
        df = aa_dict[k]
        if k == "List 9":
            df = df[
                df[index_date_key["ref_date"]] >= _remove_years(index_date_key["years"])
            ]
        elif k == "List 10":
            df = df[
                (
                    df[index_date_key["ref_date"]]
                    >= _remove_years(index_date_key["years"])
                ).any(axis=1)
            ]
        else:
            df = df[
                (
                    df[index_date_key["ref_date"]]
                    >= _remove_years(index_date_key["years"])
                )
                | (df[index_date_key["ref_date"]].isnull())
            ]
        aa_dict[k] = df
    return aa_dict


def convert_dates(aa_dict, dates):
    for k in aa_dict.keys():
        df = aa_dict[k]
        for date_field in dates[k]:
            df[date_field] = pd.to_datetime(df[date_field], format="%d/%m/%Y").dt.date
        aa_dict[k] = df
    return aa_dict


def export_file(output, aa_dict):
    output_path = Path(output, f"AnnexA_merged.xlsx")
    with pd.ExcelWriter(output_path) as writer:
        for k in aa_dict.keys():
            df = aa_dict[k]
            df.to_excel(writer, sheet_name=k, index=False)



import logging

from sfdata_stream_parser.filters.generic import streamfilter, pass_event

log = logging.getLogger(__name__)


@streamfilter(
    check=lambda x: x.get("column_header")
    in ["Child Unique ID", "Individual adopter identifier"],
    fail_function=pass_event,
)
def create_la_child_id(event, la_code):
    """
    Creates an identifier from a combination of the Child Unique ID and Local Authority so matching child IDs
    are not removed in the merging a de-duping steps

    :param event: A filtered list of event objects
    :param la_code: The 3-character LA code used to identify a local authority
    :return: An updated list of event objects
    """
    la_child_id = f"{event.value}_{la_code}"
    yield event.from_event(event, value=la_child_id)



import logging

log = logging.getLogger(__name__)


def to_integer(value):
    """
    Convert any strings that should be integers based on the config into integers
    :param value: Some value to convert to an integer
    :return: Either an integer value or a blank string
    """
    if isinstance(value, str) and value[-2:] == ".0":
        return int(float(value))
    if value or value == 0:
        return int(value)
    else:
        return ""



import logging
from typing import List

import xmlschema
from sfdata_stream_parser.checks import type_check
from sfdata_stream_parser import events
from sfdata_stream_parser.collectors import collector, block_check
from sfdata_stream_parser.filters.generic import streamfilter, pass_event
from xmlschema import XMLSchemaValidatorError

log = logging.getLogger(__name__)


@streamfilter(default_args=lambda: {"context": []})
def add_context(event, context: List[str]):
    """
    Adds 'context' to XML structures. For each :class:`sfdata_stream_parser.events.StartElement` the tag name is
    added to a 'context' tuple, and for each :class:`sfdata_stream_parser.events.EndElement` the context is popped.

    For all other events, the context tuple is set as-is.

    Provides: context
    """
    if isinstance(event, events.StartElement):
        context.append(event.tag)
        local_context = tuple(context)
    elif isinstance(event, events.EndElement):
        local_context = tuple(context)
        context.pop()
    else:
        local_context = tuple(context)

    return event.from_event(event, context=local_context)


@streamfilter(check=type_check(events.TextNode), fail_function=pass_event)
def strip_text(event):
    """
    Strips surrounding whitespaces from :class:`sfdata_stream_parser.events.TextNode`. If the event does
    not have a text property then this filter fails silently.

    If there is no content at all, then the node is not returned.
    """
    if not hasattr(event, "text"):
        return event

    if event.text is None:
        return None

    text = event.text.strip()
    if len(text) > 0:
        return event.from_event(event, text=text)
    else:
        return None


@streamfilter()
def add_schema(event, schema: xmlschema.XMLSchema):
    """
    Requires each event to have event.context as set by :func:`add_context`

    Based on the context (a tuple of element tags) it will set path which is the
    derived path (based on the context tags) joined by '/' and schema holding the
    corresponding schema element, if found.

    Provides: path, schema
    """
    assert (
        event.context
    ), "This filter required event.context to be set - see add_context"
    path = "/".join(event.context)
    tag = event.context[-1]
    el = schema.get_element(tag, path)
    return event.from_event(event, path=path, schema=el)


def _get_validation_error(schema, node) -> XMLSchemaValidatorError:
    try:
        schema.validate(node)
        return None
    except XMLSchemaValidatorError as e:
        return e


@streamfilter(check=type_check(events.StartElement), fail_function=pass_event)
def validate_elements(event):
    """
    Validates each element, and if not valid, sets the properties:

    * valid - (always False)
    * validation_message - a descriptive validation message
    """
    validation_error = _get_validation_error(event.schema, event.node)
    if validation_error is None:
        return event

    message = (
        validation_error.reason
        if hasattr(validation_error, "reason")
        else validation_error.message
    )
    return events.StartElement.from_event(
        event, valid=False, validation_message=message
    )


@streamfilter(check=type_check(events.StartElement), fail_function=pass_event)
def prop_to_attribute(event, prop_name):
    """
    Elevates an event property to an XML attribute.
    """
    if hasattr(event, prop_name):
        attrs = getattr(event, "attrs", {})
        attrs[prop_name] = getattr(event, prop_name)
        return events.StartElement.from_event(event, attrs=attrs)
    else:
        return event


@collector(check=block_check(events.StartElement), receive_stream=True)
def remove_invalid(stream, tag_name):
    """
    Filters out events with the given tag name if they are not valid
    """
    stream = list(stream)
    first = stream[0]
    last = stream[-1]
    stream = stream[1:-1]

    if first.tag == tag_name and not getattr(first, "valid", True):
        yield from []
    else:
        yield first

        if len(stream) > 0:
            yield from remove_invalid(stream, tag_name=tag_name)

        yield last


@streamfilter(check=lambda x: True)
def counter(event, counter_check, context):
    if counter_check(event):
        context["pass"] += 1
    else:
        context["fail"] += 1
    return event



import logging

from sfdata_stream_parser.filters.generic import streamfilter, pass_event

from liiatools.datasets.shared_functions.converters import (
    to_short_postcode,
    to_month_only_dob,
)

log = logging.getLogger(__name__)


@streamfilter(
    check=lambda x: x.get("header") in ["HOME_POST", "PL_POST"],
    fail_function=pass_event,
)
def degrade_postcodes(event):
    """
    Convert all values that should be postcodes to shorter postcodes

    :param event: A filtered list of event objects of type Cell
    :return: An updated list of event objects
    """
    text = to_short_postcode(event.cell)
    return event.from_event(event, cell=text)


@streamfilter(
    check=lambda x: x.get("header") in ["DOB", "MC_DOB"], fail_function=pass_event
)
def degrade_dob(event):
    """
    Convert all values that should be dates of birth to months and year of birth

    :param event: A filtered list of event objects of type Cell
    :return: An updated list of event objects
    """
    text = to_month_only_dob(event.cell)
    return event.from_event(event, cell=text)


def degrade(stream):
    """
    Compile the degrading functions
    """
    stream = degrade_postcodes(stream)
    stream = degrade_dob(stream)
    return stream



import os
import pandas as pd
import numpy as np
from liiatools.datasets.social_work_workforce.SWFtools.dataprocessing.converter import (
    ORG_ROLE_DICT,
    SENIORITY_CODE_DICT,
)
import liiatools.datasets.social_work_workforce.SWFtools.util.work_path as work_path


def seniority():
    # ===== Read file ===== #
    file = "merged_modified.csv"
    path = work_path.flatfile_folder
    requestPath = work_path.request
    pathFile = os.path.join(path, file)
    df = pd.read_csv(pathFile)

    df = df[
        [
            "YearCensus",
            "AgencyWorker",
            "SWENo",
            "RoleStartDate",
            "RoleEndDate",
            "OrgRole",
        ]
    ]

    listNew = []
    df = df.reset_index()  # make sure indexes pair with number of rows
    for index, row in df.iterrows():
        if row["RoleStartDate"] == row["YearCensus"]:
            listNew.append("New")
        else:
            listNew.append("Not")

    df["NewOrNot"] = listNew

    df = df[
        [
            "YearCensus",
            "AgencyWorker",
            "SWENo",
            "RoleStartDate",
            "RoleEndDate",
            "OrgRole",
            "NewOrNot",
        ]
    ]

    listLeft = []
    df = df.reset_index()  # make sure indexes pair with number of rows
    for index, row in df.iterrows():
        if row["RoleEndDate"] == row["YearCensus"]:
            listLeft.append("Left")
        else:
            listLeft.append("Not")

    df["LeftOrNot"] = listLeft

    df = df[
        [
            "YearCensus",
            "AgencyWorker",
            "SWENo",
            "RoleStartDate",
            "RoleEndDate",
            "OrgRole",
            "NewOrNot",
            "LeftOrNot",
        ]
    ]

    listSen = []
    df = df.reset_index()  # make sure indexes pair with number of rows
    for index, row in df.iterrows():
        if row["RoleStartDate"] == row["YearCensus"]:
            listSen.append(1)
        elif row["AgencyWorker"] == 1:
            listSen.append(5)
        elif row["OrgRole"] == 5 or row["OrgRole"] == 6:
            listSen.append(2)
        elif row["OrgRole"] == 2 or row["OrgRole"] == 3 or row["OrgRole"] == 4:
            listSen.append(3)
        elif row["OrgRole"] == 1:
            listSen.append(4)

    df["SeniorityCode"] = listSen

    df = df[
        [
            "YearCensus",
            "SWENo",
            "RoleStartDate",
            "NewOrNot",
            "RoleEndDate",
            "LeftOrNot",
            "AgencyWorker",
            "OrgRole",
            "SeniorityCode",
        ]
    ]

    fileOutN = "Seniority.csv"
    requestPath = work_path.request
    fileOut = os.path.join(requestPath, fileOutN)
    df.to_csv(fileOut, index=False)


def seniority_forecast_04():
    # ===== Read file ===== #
    file = "FTESum_2020.xlsx"
    requestPath = work_path.request
    pathFile = os.path.join(requestPath, file)
    dfSen = pd.read_excel(pathFile)

    # ===== Rename column ===== #
    dfSen.rename(columns={"FTESum": "2020"}, inplace=True)

    # ===== Read file ===== #
    file = "population_growth_table.xlsx"
    requestPath = work_path.request
    pathFile = os.path.join(requestPath, file)
    p_df = pd.read_excel(pathFile)

    countYearBefore = 2019
    countYearNext = 2020
    for count in range(5):
        countYearBefore = countYearBefore + 1
        countYearNext = countYearNext + 1
        # Havering
        dfSen.loc[dfSen["LEAName"] == "Havering", str(countYearNext)] = (
            dfSen[str(countYearBefore)] / p_df.loc[0, str(countYearBefore)]
        ) * p_df.loc[0, str(countYearNext)]
        # Barking and Dagenham
        dfSen.loc[dfSen["LEAName"] == "Barking and Dagenham", str(countYearNext)] = (
            dfSen[str(countYearBefore)] / p_df.loc[1, str(countYearBefore)]
        ) * p_df.loc[1, str(countYearNext)]
        # Redbridge
        dfSen.loc[dfSen["LEAName"] == "Redbridge", str(countYearNext)] = (
            dfSen[str(countYearBefore)] / p_df.loc[2, str(countYearBefore)]
        ) * p_df.loc[2, str(countYearNext)]
        # Newham
        dfSen.loc[dfSen["LEAName"] == "Newham", str(countYearNext)] = (
            dfSen[str(countYearBefore)] / p_df.loc[3, str(countYearBefore)]
        ) * p_df.loc[3, str(countYearNext)]
        # Tower Hamlets
        dfSen.loc[dfSen["LEAName"] == "Tower Hamlets", str(countYearNext)] = (
            dfSen[str(countYearBefore)] / p_df.loc[4, str(countYearBefore)]
        ) * p_df.loc[4, str(countYearNext)]
        # Waltham Forest
        dfSen.loc[dfSen["LEAName"] == "Waltham Forest", str(countYearNext)] = (
            dfSen[str(countYearBefore)] / p_df.loc[5, str(countYearBefore)]
        ) * p_df.loc[5, str(countYearNext)]

    dfSen["2020"] = dfSen["2020"].round(3)
    dfSen["2021"] = dfSen["2021"].round(3)
    dfSen["2022"] = dfSen["2022"].round(3)
    dfSen["2023"] = dfSen["2023"].round(3)
    dfSen["2024"] = dfSen["2024"].round(3)
    dfSen["2025"] = dfSen["2025"].round(3)

    dfSen = dfSen.drop(["YearCensus"], axis=1)

    # ===== Save and export file ===== #
    fileOutN = "seniority_forecast_04_clean.xlsx"
    requestPath = work_path.request
    fileOut = os.path.join(requestPath, fileOutN)
    dfSen.to_excel(fileOut, index=False, merge_cells=False)


def seniority_forecast_5c():
    # ===== Read file ===== #
    file = "Seniority.csv"
    path = work_path.flatfile_folder
    requestPath = work_path.request
    pathFile = os.path.join(requestPath, file)
    dfSen = pd.read_csv(pathFile)

    dfSen["OrgRoleName"] = dfSen.OrgRole.map(
        {int(key): ORG_ROLE_DICT[key] for key in ORG_ROLE_DICT}
    )
    dfSen["SeniorityName"] = dfSen.SeniorityCode.map(
        {int(key): SENIORITY_CODE_DICT[key] for key in SENIORITY_CODE_DICT}
    )

    dfSen = dfSen[
        [
            "YearCensus",
            "SWENo",
            "RoleStartDate",
            "NewOrNot",
            "RoleEndDate",
            "LeftOrNot",
            "AgencyWorker",
            "OrgRole",
            "OrgRoleName",
            "SeniorityCode",
            "SeniorityName",
        ]
    ]

    # ===== Save and export file ===== #
    fileOutN = "SeniorityComp.csv"
    requestPath = work_path.request
    fileOut = os.path.join(requestPath, fileOutN)
    dfSen.to_csv(fileOut, index=False)

    # ===== Sort values ===== #
    dfSen = dfSen.sort_values(by=["SWENo", "YearCensus"])

    # ============================================================================================ #

    # ===== Read file ===== #
    file = "merged_modified.csv"
    path = work_path.flatfile_folder
    requestPath = work_path.request
    pathFile = os.path.join(path, file)
    dfMerged = pd.read_csv(pathFile)

    dfMerged = dfMerged.sort_values(by=["SWENo", "YearCensus"])

    dfMerged["OrgRole"] = dfSen["OrgRole"]
    dfMerged["OrgRoleName"] = dfSen["OrgRoleName"]
    dfMerged["SeniorityCode"] = dfSen["SeniorityCode"]
    dfMerged["SeniorityName"] = dfSen["SeniorityName"]

    # ===== Save and export file ===== #
    fileOutN = "CompMergSen.csv"
    requestPath = work_path.request
    fileOut = os.path.join(requestPath, fileOutN)
    dfMerged.to_csv(fileOut, index=False)

    # ================================================================================================================= #


def progressed():
    # ===== Read file ===== #
    file = "Seniority.csv"
    requestPath = work_path.request
    pathSen = os.path.join(requestPath, file)
    df = pd.read_csv(pathSen)
    df = df.sort_values(by=["SWENo", "YearCensus"])

    df["Progress"] = np.where(
        df["SWENo"] == df["SWENo"].shift(),
        np.where(
            df["SeniorityCode"] == df["SeniorityCode"].shift(), "No", "Progressed"
        ),
        "Unknown",
    )

    fileOutN = "CompletProgressed.csv"
    requestPath = work_path.request
    fileOut = os.path.join(requestPath, fileOutN)
    df.to_csv(fileOut, index=False)



"""Main script where all the others will be called"""

import time
import liiatools.datasets.social_work_workforce.SWFtools.util.AppLogs as AppLog
import liiatools.datasets.social_work_workforce.SWFtools.util.work_path as work_path
import liiatools.datasets.social_work_workforce.SWFtools.dataprocessing.file_operations as fop
from liiatools.datasets.social_work_workforce.SWFtools.analysis.growth_tables import (
    growth_tables,
)
from liiatools.datasets.social_work_workforce.SWFtools.analysis.pivotGen import pivotGen
from liiatools.datasets.social_work_workforce.SWFtools.analysis.seniority import (
    seniority,
    progressed,
    seniority_forecast_5c,
    seniority_forecast_04,
)
from liiatools.datasets.social_work_workforce.SWFtools.analysis.FTESum import (
    FTESum,
    FTESum_2020,
)

start = time.time()

# Initialising logger (present throughout the program)
AppLog.initialise()

# CONVERSION STEPS
# Recreate all folders present in 'cin' into 'flatfiles' adding a 'la_log' folder to store all runtime logs
work_path.check_flatfiles_folder()

# Validate, process, and convert all XML files found in LA directories to CSV then merge them
fop.process_all_input_files()

# For creating spreadsheets, tables and other files that will be stored in the request folder


# Outputs a file of extension 'xlsx' named 'growth_tables' (does not process any file, contains hardcoded values)
growth_tables()
# Creates a pivot table grouping data on Census year, age, gender, and LEA name; and summing on FTE and SWE
pivotGen()

seniority()
progressed()
seniority_forecast_5c()
FTESum()
FTESum_2020()
seniority_forecast_04()

end = time.time()
total_time = round(end - start, 3)

print()
print("Execution Time:", total_time)

AppLog.log_footer(total_time)



from collections import Counter


def consume(stream) -> Counter:
    """
        Ensures the stream is consumed and returns a summary of the numbers of each event that has been encountered
    """
    stream_types = [type(ev) for ev in stream]
    return Counter(stream_types)





"""This file contains constants and functions used in converting and merging XML data from CSWW folder"""

import hashlib
import datetime as dt
from typing import Final, Dict, List
from decouple import config


# === CONSTANTS === #
LEA_DICT: Final[Dict[str, str]] = {
    "301": "Barking and Dagenham",
    "316": "Newham",
    "317": "Redbridge",
    "211": "Tower Hamlets",
    "311": "Havering",
    "320": "Waltham Forest",
}

ETHNICITY_GROUP_DICT: Final[Dict[str, str]] = {
    "WBRI": "White - British",
    "WIRI": "White - Irish",
    "WOTH": "Any Other White Background",
    "MWBC": "White and Black Caribbean",
    "MWBA": "White and Black African",
    "MWAS": "White and Asian",
    "MOTH": "Any Other Mixed background",
    "AIND": "Indian",
    "APKN": "Pakistani",
    "ABAN": "Bangladeshi",
    "AOTH": "Any Other Asian Background",
    "BCRB": "Black Caribbean",
    "BAFR": "Black African",
    "BOTH": "Any Other Black Background",
    "CHNE": "Chinese",
    "OOTH": "Any Other Ethnic Group",
    "REFU": "Declared not stated or Refused",
    "NOBT": "Information Not Yet Obtained",
}

ETHNICITY_DICT: Final[Dict[str, str]] = {
    "WBRI": "White - British",
    "WIRI": "White - Irish",
    "WOTH": "Any Other White Background",
    "MWBC": "White and Black Caribbean",
    "MWBA": "White and Black African",
    "MWAS": "White and Asian",
    "MOTH": "Any Other Mixed background",
    "AIND": "Indian",
    "APKN": "Pakistani",
    "ABAN": "Bangladeshi",
    "AOTH": "Any Other Asian Background",
    "BCRB": "Black Caribbean",
    "BAFR": "Black African",
    "BOTH": "Any Other Black Background",
    "CHNE": "Chinese",
    "OOTH": "Any Other Ethnic Group",
    "REFU": "Declared not stated or Refused",
    "NOBT": "Information Not Yet Obtained",
}

GENDER_DICT: Final[Dict[str, str]] = {
    "0": "Not known (gender has not been recorded)",
    "1": "male",
    "2": "female",
    "9": "not specified (indeterminate; unable to be classified as either male or female)",
}

ORG_ROLE_DICT: Final[Dict[str, str]] = {
    "1": "Senior Manager",
    "2": "Middle Manager",
    "3": "First Line Manager",
    "4": "Senior Practitioner",
    "5": "Case Holder",
    "6": "Qualified without cases",
}

SENIORITY_CODE_DICT: Final[Dict[str, str]] = {
    "1": "Newly qualified",
    "2": "Early career",
    "3": "Experienced",
    "4": "Senior",
    "5": "Agency",
}

QUAL_LEVEL_DICT: Final[Dict[str, str]] = {
    "1": "Undergraduate",
    "2": "Postgraduate",
    "3": "Other (for example any other qualification)",
}


# === FUNCTIONS === #
def swe_hash(worker: Dict[str, str]):
    """
    Converts the **SWENo** field to a hash code represented in HEX
    :param worker: A dictionary containing worker data
    :return: None
    """
    swe_num = worker["SWENo"]

    private_string = config("sec_str", default="")

    private_key = swe_num + private_string

    # Preparing plain text (SWENo) to hash it
    plaintext = private_key.encode()

    hash_algorithm = hashlib.sha3_256(plaintext)

    worker["SWENo"] = hash_algorithm.hexdigest()


def convert_dates(worker: Dict[str, str]):
    """
    This function will:
    1) Add and compute the value for the following columns: **Age**, **RoleStartDate**, **RoleEndDate**, **DoB_year**
    (all of these values contain only their respective year, not the full date)
    2) Remove the **PersonBirthDate** field inside each worker dictionary
    :param worker: A dictionary containing worker data
    :return: None
    """
    date_format = "%Y-%m-%d"

    if "PersonBirthDate" in worker:
        birth_date = worker["PersonBirthDate"]
        worker["DoB_year"] = str(dt.datetime.strptime(birth_date, date_format).year)
        worker["Age"] = str(
            dt.datetime.today().year
            - dt.datetime.strptime(birth_date, date_format).year
        )

        del worker["PersonBirthDate"]

    if "RoleStartDate" in worker:
        worker["RoleStartDate"] = str(
            dt.datetime.strptime(worker["RoleStartDate"], date_format).year
        )

    if "RoleEndDate" in worker:
        worker["RoleEndDate"] = str(
            dt.datetime.strptime(worker["RoleEndDate"], date_format).year
        )


def column_transfer(workers: List[Dict[str, str]], lea: str, year_census: str):
    """
    Adds new columns to the table.
    :param workers: A list of dictionaries containing worker data
    :param lea: Local Authority code
    :param year_census: Census year
    :return: None
    """
    for worker in workers:
        if lea in LEA_DICT:
            worker["LEAName"] = LEA_DICT[lea]
        else:
            worker["LEAName"] = ""

        # Adding ethnicity extra columns
        if "Ethnicity" in worker:
            ethnicity = worker["Ethnicity"]
            worker["Ethnicity_Group"] = ETHNICITY_GROUP_DICT[ethnicity]
            worker["Ethnicity_Compact"] = ETHNICITY_DICT[ethnicity]
        else:
            worker["Ethnicity_Group"] = ""
            worker["Ethnicity_Compact"] = ""

        # Adding gender extra column
        if "GenderCurrent" in worker:
            worker["Gender"] = GENDER_DICT[worker["GenderCurrent"]]
        else:
            worker["Gender"] = ""

        # Organisation role extra column
        if "OrgRole" in worker:
            worker["OrgRoleName"] = ORG_ROLE_DICT[worker["OrgRole"]]
        else:
            worker["OrgRoleName"] = ""

        # Qualification level
        if "QualLevel" in worker:
            worker["QualLevelName"] = QUAL_LEVEL_DICT[worker["QualLevel"]]
        else:
            worker["QualLevelNAme"] = ""

        worker["YearCensus"] = year_census
        worker["LEA"] = lea



import tablib
import logging
from pathlib import Path
from datetime import datetime

from sfdata_stream_parser import events

log = logging.getLogger(__name__)


def parse_csv(input):
    """
    Parse the csv and return the row number, column number, header name and cell value

    :param input: Location of file to be cleaned
    :return: List of event objects containing filename, header and cell information
    """
    filename = str(Path(input).resolve().stem)

    with open(input, "rt") as f:
        yield events.StartContainer(filename=filename)
        data = tablib.import_set(f, format="csv")
        yield events.StartTable(filename=filename, headers=data.headers)
        for r_ix, row in enumerate(data):
            yield events.StartRow(filename=filename)
            for c_ix, cell in enumerate(row):
                yield events.Cell(
                    filename=filename,
                    r_ix=r_ix,
                    c_ix=c_ix,
                    header=data.headers[c_ix],
                    cell=cell,
                )
            yield events.EndRow(filename=filename)
        yield events.EndTable(filename=filename)
        yield events.EndContainer()



from pathlib import Path
import pandas as pd
import logging

log = logging.getLogger(__name__)


def read_file(file):
    """
    Reads the csv file as a pandas DataFrame
    """
    filepath = Path(file)
    s903_df = pd.read_csv(filepath, index_col=None)
    return s903_df


def match_load_file(s903_df, column_names):
    """
    Matches the columns in the DataFrame against one of the 10 SSDA903 file types
    """
    for table_name, expected_columns in column_names.items():
        if set(s903_df.columns) == set(expected_columns):
            return table_name


def merge_la_files(output, s903_df, table_name):
    """
    Looks for existing file of the same type and merges with new file if found
    """
    old_file = Path(output, f"SSDA903_{table_name}_merged.csv")
    if old_file.is_file():
        old_df = pd.read_csv(old_file, index_col=None)
        merged_df = pd.concat([s903_df, old_df], axis=0)
    else:
        merged_df = s903_df
    return merged_df


def convert_datetimes(s903_df, dates, table_name):
    """
    Ensures that all date fields have been parsed as dates
    """
    for date_field in dates[table_name]:
        s903_df[date_field] = pd.to_datetime(s903_df[date_field], format="%Y/%m/%d")
    return s903_df


def deduplicate(s903_df, table_name, sort_order, dedup):
    """
    Sorts and removes duplicate records from merged files following schema
    """
    s903_df = s903_df.sort_values(
        sort_order[table_name], ascending=False, ignore_index=True
    )
    s903_df = s903_df.drop_duplicates(subset=dedup[table_name], keep="first")
    return s903_df


def remove_old_data(s903_df, num_of_years, new_year_start_month, as_at_date):
    """
    Removes data older than a specified number of years as at reference date

    :param s903_df: Dataframe containing csv data
    :param num_of_years: The number of years to go back
    :param new_year_start_month: The month which signifies start of a new year for data retention policy
    :param as_at_date: The reference date against which we are checking the valid range
    :return: Dataframe with older years removed
    """
    current_year = pd.to_datetime(as_at_date).year
    current_month = pd.to_datetime(as_at_date).month

    if current_month < new_year_start_month:
        earliest_allowed_year = current_year - num_of_years
    else:
        earliest_allowed_year = current_year - num_of_years + 1  # roll forward one year

    s903_df = s903_df[s903_df["YEAR"] >= earliest_allowed_year]
    return s903_df


def convert_dates(s903_df, dates, table_name):
    """
    Ensures that all date fields have been parsed as dates
    """
    for date_field in dates[table_name]:
        s903_df[date_field] = pd.to_datetime(
            s903_df[date_field], format="%Y/%m/%d"
        ).dt.date
    return s903_df


def export_la_file(output, table_name, s903_df):
    """
    Writes the output as a csv
    """
    output_path = Path(output, f"SSDA903_{table_name}_merged.csv")
    s903_df.to_csv(output_path, index=False)



"""Functions for validating XML files of LAs and getting validated worker data"""

from typing import List, Dict, Final

import lxml.etree as etree

import xmlschema
from xmlschema import XMLSchemaValidationError
from liiatools.datasets.social_work_workforce.SWFtools.util.work_path import XML_SCHEMA

from liiatools.datasets.social_work_workforce.SWFtools.dataprocessing.validation.validation_error import (
    ValidationError,
    ERROR_CAUSE,
)

CIN_XML_SCHEMA: Final = xmlschema.XMLSchema11(XML_SCHEMA)
WORKER_SCHEMA: Final = CIN_XML_SCHEMA.find("//CSWWWorker")
NON_AGENCY_MANDATORY_TAG: Final = [
    "PersonBirthDate",
    "GenderCurrent",
    "Ethnicity",
    "QualInst",
    "StepUpGrad",
    "OrgRole",
    "RoleStartDate",
    "StartOrigin",
    "FTE30",
    "WorkingDaysLost",
    "FrontlineGrad",
]


def convert_schema_error(schema_error: XMLSchemaValidationError) -> ValidationError:
    """
    Converts an **XMLSchemaValidationError** to a custom validation error. See *'validation_error.py'*
    """

    # Structural error
    # For this error view source file xmlschema/validators/groups.py (at line 1266)
    if schema_error.validator.__class__ == xmlschema.validators.Xsd11Group:
        # tuple (minOccurs, maxOccurs) - these numbers are specified inside the XML schema
        occurs_range = schema_error.args[3].occurs

        # An <int> representing how many times the tag that is being validated occurs
        # This value is 0 if there is a missmatch between the tag that was expected and the
        # tag that was actually read

        occurs = schema_error.args[4]

        # The tag of the read element
        read_tag = schema_error.elem[schema_error.args[2]].tag

        # The tag of the expected element
        expected_tag = schema_error.args[3].name

        if read_tag != expected_tag:
            return ValidationError(
                ERROR_CAUSE.GROUP_FORMAT, schema_error.elem.tag, schema_error.sourceline
            )
        elif occurs < occurs_range[0]:
            return ValidationError(
                ERROR_CAUSE.MISSING_TAG, schema_error.elem.tag, schema_error.sourceline
            )
        elif occurs > occurs_range[1]:
            return ValidationError(
                ERROR_CAUSE.TOO_MANY_TAGS,
                schema_error.elem.tag,
                schema_error.sourceline,
            )
        else:
            print(f"UNEXPECTED ERROR!\n{schema_error}")
            return None
    # Value does not match to any enumerated value
    elif schema_error.validator.__class__ == xmlschema.validators.XsdEnumerationFacets:
        return ValidationError(
            ERROR_CAUSE.VALUE, schema_error.elem.tag, schema_error.sourceline
        )
    # Value does not match pattern specified
    elif schema_error.validator.__class__ == xmlschema.validators.XsdPatternFacets:
        return ValidationError(
            ERROR_CAUSE.VALUE, schema_error.elem.tag, schema_error.sourceline
        )
    # Value is of unexpected type
    elif schema_error.validator.__class__ == xmlschema.validators.XsdAtomicBuiltin:
        return ValidationError(
            ERROR_CAUSE.TYPE, schema_error.elem.tag, schema_error.sourceline
        )
    # Value is below accepted range minimum inclusive
    elif schema_error.validator.__class__ == xmlschema.validators.XsdMinInclusiveFacet:
        return ValidationError(
            ERROR_CAUSE.MIN_RANGE, schema_error.elem.tag, schema_error.sourceline
        )
    # Value is above accepted range maximum inclusive
    elif schema_error.validator.__class__ == xmlschema.validators.XsdMaxInclusiveFacet:
        return ValidationError(
            ERROR_CAUSE.MAX_RANGE, schema_error.elem.tag, schema_error.sourceline
        )


def is_acceptable(parsed_xml: etree.Element, error_list: List[ValidationError]) -> bool:
    """
    Validates **'Header'** and **'LALevelVacancies'** tags then checks that the format of the message is correct:
    **'Header'**, **'LALevelVacancies'** followed by **'CSWWWorker'** n times

    :param parsed_xml: The root element of the xml document to be checked.
    :param error_list: A list where all validation errors are placed into.
    :return: Returns 'True' if the document can be further processed of 'False' otherwise.
    """

    if not __is_message_format_valid(parsed_xml, error_list):
        return False

    la_vacancies_schema = CIN_XML_SCHEMA.find("//LALevelVacancies")
    la_vacancies_xml = parsed_xml.xpath("//LALevelVacancies")[0]

    header_schema = CIN_XML_SCHEMA.find("//Header")
    header_xml = parsed_xml.xpath("//Header")[0]

    verr = [
        convert_schema_error(error) for error in header_schema.iter_errors(header_xml)
    ]
    verr.extend(
        [
            convert_schema_error(error)
            for error in la_vacancies_schema.iter_errors(la_vacancies_xml)
        ]
    )

    if len(verr) != 0:
        error_list.extend(verr)
        return False

    return True


# Shallow validation of 'Header', 'LALevelVacancies', and 'CSWWWorker' tags
def __is_message_format_valid(
    parsed_xml: etree.ElementTree, error_list: List[ValidationError]
) -> bool:
    can_continue = True

    for error in CIN_XML_SCHEMA.iter_errors(parsed_xml, max_depth=1):
        va = convert_schema_error(error)

        # A group format means that the Header, LAVacancies and CSWWWorker tags are just out of order, can still proceed
        if va is not None and va.cause != ERROR_CAUSE.GROUP_FORMAT:
            can_continue = False

        error_list.append(va)

    return can_continue


def get_valid_worker_data(
    worker_element: etree.Element, error_list: List[ValidationError]
) -> Dict[str, str]:
    """
    Takes in a worker element (lxml.etree), validates it and returns the worker data as a dictionary
    or **'None'** if the worker data is invalid.

    :param worker_element: A tree element (lxml.etree) that has the tag 'CSWWWorker' .
    :param error_list: A list where all validation errors are placed into.
    :return: A dictionary that contains the worker data. Child tags are the keys and
    their value as the dictionary's value
    """

    # === XML SCHEMA VALIDATION ===
    va: List[ValidationError] = list(
        map(convert_schema_error, WORKER_SCHEMA.iter_errors(worker_element))
    )

    # === BUILDING WORKER DATA SET; FURTHER VALIDATION ===
    worker_data = {}

    for elem in worker_element:
        worker_data[elem.tag] = elem.text

    is_agency_worker = worker_data["AgencyWorker"] == "1"
    is_leaver = "RoleEndDate" in worker_data
    is_starter = "RoleStartDate" in worker_data

    # Validation rules if the worker is a non-agency worker
    if not is_agency_worker:
        # If a worker is a non-agency worker, check that mandatory tags are present
        for tag in NON_AGENCY_MANDATORY_TAG:
            if tag not in worker_data:
                va.append(
                    ValidationError(
                        ERROR_CAUSE.NON_AGENCY_MANDATORY_MISSING,
                        tag,
                        worker_element.sourceline,
                    )
                )

    # Validation rules if the worker is a starter (condition checked via eXclusive OR)
    if is_starter ^ ("StartOrigin" in worker_data):
        va.append(
            ValidationError(
                ERROR_CAUSE.MISSING_TAG, "StartOrigin", worker_element.sourceline
            )
        )

    # Validation rules if the worker is a leaver
    if is_leaver and worker_data["FTE"] != "0":
        sourceline = worker_element.xpath("//FTE")[0].sourceline
        va.append(ValidationError(ERROR_CAUSE.LEAVER_FTE, "FTE", sourceline))

    if is_leaver ^ ("ReasonLeave" in worker_data):
        if is_leaver:
            va.append(ERROR_CAUSE.MISSING_TAG, "ReasonLeave", worker_element.sourceline)
        else:
            va.append(
                ValidationError(
                    ERROR_CAUSE.LEAVER_UNEXPECTED,
                    "ReasonLeave",
                    worker_element.sourceline,
                )
            )

    if is_leaver ^ ("LeaverDestination" in worker_data):
        if is_leaver:
            va.append(
                ERROR_CAUSE.MISSING_TAG, "LeaverDestination", worker_element.sourceline
            )
        else:
            va.append(
                ValidationError(
                    ERROR_CAUSE.LEAVER_UNEXPECTED,
                    "LeaverDestination",
                    worker_element.sourceline,
                )
            )

    # If there are any errors the element is invalid, return None and move on
    if len(va) != 0:
        error_list.extend(va)
        return None

    return worker_data



import os
import pandas as pd
import liiatools.datasets.social_work_workforce.SWFtools.util.work_path as work_path


def FTESum():
    # ===== Read file ===== #
    file = "CompMergSen.csv"
    requestPath = work_path.request
    pathFile = os.path.join(requestPath, file)
    df = pd.read_csv(pathFile)

    df5C = df[["LEAName", "YearCensus", "SeniorityCode", "SeniorityName", "FTE"]]

    df5C = df5C.sort_values(by=["LEAName", "YearCensus", "SeniorityCode"])

    df5C = df.groupby(["LEAName", "YearCensus", "SeniorityCode", "SeniorityName"]).agg(
        FTESum=("FTE", "sum")
    )

    # ===== Save and export file ===== #
    fileOutN = "FTESum_5d.xlsx"
    requestPath = work_path.request
    fileOut = os.path.join(requestPath, fileOutN)
    df5C.to_excel(fileOut, merge_cells=False)


def FTESum_2020():
    # ===== Read file ===== #
    file = "CompMergSen.csv"
    requestPath = work_path.request
    pathFile = os.path.join(requestPath, file)
    df = pd.read_csv(pathFile)

    df2020 = df[df["YearCensus"] == 2020]

    df5D = df2020[["LEAName", "YearCensus", "SeniorityCode", "SeniorityName", "FTE"]]

    df5D = df2020.groupby(
        ["LEAName", "YearCensus", "SeniorityCode", "SeniorityName"]
    ).agg(FTESum=("FTE", "sum"))

    # ===== Save and export file ===== #
    fileOutN = "FTESum_2020.xlsx"
    requestPath = work_path.request
    fileOut = os.path.join(requestPath, fileOutN)
    df5D.to_excel(fileOut, merge_cells=False)



import sys
import os
import liiatools.datasets.social_work_workforce.SWFtools.util.AppLogs as AppLogs

# Setting the paths to work and executing and using the scripts

# Define filepaths
main_folder = r"/workspaces/liia-tools/liiatools/spec/social_work_workforce"

# CSWW files must be in one "LA" folder per LA, in the cssw_folder
csww_folder = os.path.join(main_folder, "samples/csww")
# print (f"csww_folder is {csww_folder}")

# Flat files
flatfile_folder = os.path.join(main_folder, "samples/flatfiles")

# Outputs
output_folder = os.path.join(main_folder, "samples/outputs")

# Chris requests folder
request = os.path.join(main_folder, "samples/request")

# Workforce XML schema
XML_SCHEMA = os.path.join(main_folder, "social_work_workforce_2022.xsd")

# Runtime log files
runtime_log_files = os.path.join(main_folder, "samples/log_files")

# Local Authority directories
la_directories = [folder for folder in os.scandir(csww_folder) if os.path.isdir(folder)]


def check_flatfiles_folder():
    """
    Checks that the 'flatfiles' folder contains all directories that are present in 'csww' folder and creates them if
    they are not present.
    :return: None
    """
    total_dirs = len(la_directories)
    checked_dirs = 0
    AppLogs.log(f"Checking {total_dirs} directories...", console_output=True)

    for directory in la_directories:
        la_flat_files_directory = os.path.join(flatfile_folder, directory.name)

        if os.path.exists(la_flat_files_directory):
            AppLogs.log(
                f"{directory.name} - Directory already exists", console_output=True
            )
        # If the folder does not exist, create it and add a log folder called 'la_log'
        else:
            try:
                os.mkdir(la_flat_files_directory)

                la_log_file_directory = os.path.join(la_flat_files_directory, "la_log")
                os.mkdir(la_log_file_directory)

                AppLogs.log(f"Created directory: {directory}", console_output=True)
            except PermissionError as pe:
                AppLogs.log(f"PERMISSION ERROR!\n{pe}", console_output=True)

                print("")
                sys.exit("The program closed after error occurred at runtime.")

        checked_dirs = checked_dirs + 1
        AppLogs.log(f"Checked {checked_dirs} of {total_dirs} directories.")


def is_la_directory(name: str) -> bool:
    for d in la_directories:
        if d.name == name:
            return True

    return False



import functools
import tablib
import logging
import os

from sfdata_stream_parser import events

log = logging.getLogger(__name__)


class RowEvent(events.ParseEvent):
    pass


def coalesce_row(stream):
    """
    Create a list of the cell values for a whole row

    :param stream: The stream to output
    :return: Updated stream
    """
    row = None
    for event in stream:
        if isinstance(event, events.StartRow):
            row = []
        elif isinstance(event, events.EndRow):
            yield RowEvent.from_event(event, row=row)
            row = None
        elif (
            row is not None
            and isinstance(event, events.Cell)
            and event.header in set(event.expected_columns)
        ):
            row.append(event.cell)
        else:
            yield event


class TableEvent(events.ParseEvent):
    pass


def create_tables(stream, la_name):
    """
    Append all the rows for a given table to create one concatenated data event

    :param stream: The stream to output
    :param la_name: The name of the local authority
    :return: Updated stream
    """
    data = None
    for event in stream:
        if isinstance(event, events.StartTable):
            match_error = getattr(event, "match_error", None)
            year_error = getattr(event, "year_error", None)
            if match_error or year_error is not None:
                data = None
            else:
                data = tablib.Dataset(headers=event.expected_columns + ["LA", "YEAR"])
        elif isinstance(event, events.EndTable):
            yield event
            yield TableEvent.from_event(event, data=data)
            data = None
        elif data is not None and isinstance(event, RowEvent):
            try:
                data.append(event.row + [la_name, event.year])
            except (
                AttributeError
            ):  # raised in case event.year is missing so data is not added
                pass
        yield event


def save_tables(stream, output):
    """
    Save the data events as csv files in the Outputs directory

    :param stream: The stream to output
    :param output: The location of the output file
    :return: updated stream object.
    """
    for event in stream:
        if isinstance(event, TableEvent) and event.data is not None:
            dataset = event.data
            with open(
                f"{os.path.join(output, event.filename)}_clean.csv", "w", newline=""
            ) as f:
                f.write(dataset.export("csv"))
        yield event


def save_stream(stream, la_name, output):
    """
    Outputs stream to file

    :param stream: The stream to output
    :param la_name: Full name of the LA
    :param output: Location to write the output
    :return: Updated stream
    """
    stream = coalesce_row(stream)
    stream = create_tables(stream, la_name=la_name)
    stream = save_tables(stream, output=output)
    return stream


@functools.cache
def lookup_column_config(table_name, column_name):
    return None



from enum import Enum


class ERROR_CAUSE(Enum):
    VALUE = "Invalid value"
    TYPE = "The type of data read does not match the expected data type"
    MISSING_TAG = "Missing tag"
    TOO_MANY_TAGS = "Found more tags than expected"
    GROUP_FORMAT = "Not formatted as expected"
    MIN_RANGE = "Below the value's expected minimum value"
    MAX_RANGE = "Above the value's expected maximum value"
    NON_AGENCY_MANDATORY_MISSING = "Mandatory tag for non-agency worker missing"
    LEAVER_MANDATORY_MISSING = "Mandatory tag for leavers is missing"
    LEAVER_UNEXPECTED = "Unexpected value present, should not be present for leavers"
    LEAVER_FTE = "FTE must be 0 for leavers"


class ValidationError:
    def __init__(self, cause: ERROR_CAUSE, tag: str, sourceline: int = 0):
        self.cause = cause
        self.tag = tag
        self.sourceline = sourceline

    def __str__(self):
        return f"[Tag '{self.tag}' on line {self.sourceline}]: {self.cause.value}"



from pathlib import Path
import pandas as pd
import logging

log = logging.getLogger(__name__)


def read_file(file):
    """
    Reads the csv file as a pandas DataFrame
    """
    filepath = Path(file)
    s903_df = pd.read_csv(filepath, index_col=None)
    return s903_df


def match_load_file(s903_df, column_names):
    """
    Matches the columns in the DataFrame against one of the 10 SSDA903 file types
    """
    for table_name, expected_columns in column_names.items():
        if set(s903_df.columns) == set(expected_columns):
            return table_name


def _merge_dfs(s903_df, old_df, la_name):
    """
    Deletes existing data for new LA from pan file
    Merges new LA data to pan file
    """
    old_df = old_df.drop(old_df[old_df["LA"] == la_name].index)
    s903_df = pd.concat([s903_df, old_df], axis=0, ignore_index=True)
    return s903_df


def merge_agg_files(output, table_name, s903_df, la_name):
    """
    Checks if pan file exists
    Passes old and new file to function to be merged
    """
    output_file = Path(output, f"pan_London_SSDA903_{table_name}.csv")
    if output_file.is_file():
        old_df = pd.read_csv(output_file, index_col=None)
        s903_df = _merge_dfs(s903_df, old_df, la_name)
    return s903_df


def export_pan_file(output, table_name, s903_df):
    """
    Writes file to output directory
    """
    output_path = Path(output, f"pan_London_SSDA903_{table_name}.csv")
    s903_df.to_csv(output_path, index=False)



"""A script to write text to log files and terminal"""

import os
import platform
import sys
import datetime
from pathlib import Path
from typing import List, Final, Dict

import liiatools.datasets.social_work_workforce.SWFtools.util.work_path as work_path

log_paths: Dict[str, str] = {}

TIME_FORMAT_FILE_NAME: Final[str] = "%d_%m_%Y_%H_%M_%S"
TIME_FORMAT_LOG_ENTRY: Final[str] = "%d/%m/%Y at %H:%M:%S"

OPERATING_SYSTEM: Final = platform.uname().system


# Creates a log file for runtime logs and writes a log header into the file
def initialise():
    time_start = datetime.datetime.now()

    # The log file will be stored in a dictionary for later use
    runtime_log_file_name = time_start.strftime(TIME_FORMAT_FILE_NAME) + ".txt"
    log_paths["runtime"] = os.path.join(
        work_path.runtime_log_files, runtime_log_file_name
    )

    try:
        with open(log_paths["runtime"], "w") as file:
            file.write("=" * 20 + "\n")
            file.write(
                f"Log file created on: {time_start.strftime(TIME_FORMAT_LOG_ENTRY)}\n"
            )
            file.write(f"Operating System: {OPERATING_SYSTEM}\n")
            file.write("=" * 20 + "\n")
    except Exception as e:
        print(
            "A critical error occurred when attempting to create and write into runtime log file.\nError message:\n"
        )
        print(e)
        print("")
        sys.exit("The program closed after error occurred at runtime.")


def log_section_header(
    log_text: str, log_dir_name: str = "runtime", console_output: bool = False
):
    thick_line = "=" * 5
    log_text = f"# {thick_line} {log_text} {thick_line} #"
    log(log_text, log_dir_name, console_output)


# Writes text into a log file that corresponds to the log id passed
def log(
    log_text: str | List[str],
    log_dir_name: str = "runtime",
    console_output: bool = False,
):
    """
    Writes text into the log file of an LA or to the application's runtime log file.
    :param log_text: A line or list of lines of text that will be written in the log file.
    :param log_dir_name: Directory name of the LA as it appears in the CIN folder. This will be used as a log id.
    If no value is passed it will write to the applications log file ('runtime')

    :param console_output: If it is set to 'True' then the passed text will also be printed to console. 'False' by default.
    :return: None
    """
    time_stamp = datetime.datetime.now()
    entry_time_stamp = time_stamp.strftime(TIME_FORMAT_LOG_ENTRY)

    try:
        # If a log file was created then it's id can be found inside log_files dictionary as a key
        if log_dir_name in log_paths:
            if console_output:
                __write_to_log_file_verbose(
                    log_text, entry_time_stamp, "a", log_paths[log_dir_name]
                )
            else:
                __write_to_log_file(
                    log_text, entry_time_stamp, "a", log_paths[log_dir_name]
                )

        # If log_id was not found in log_files dictionary, check if it matches to any LA directories and create
        # it in its respective directory
        elif work_path.is_la_directory(log_dir_name):
            file_path = os.path.join(
                work_path.flatfile_folder,
                log_dir_name,
                "la_log",
                time_stamp.strftime(TIME_FORMAT_FILE_NAME) + ".txt",
            )
            log_paths[log_dir_name] = file_path

            if console_output:
                __write_to_log_file_verbose(
                    log_text, entry_time_stamp, "w", log_paths[log_dir_name]
                )
            else:
                __write_to_log_file(
                    log_text, entry_time_stamp, "w", log_paths[log_dir_name]
                )

    except Exception as e:
        print("An error occurred when writing into the log file.\nError message:\n")
        print(e)


def __write_to_log_file(
    log_text: str | List[str], entry_time_stamp: str, mode: str, path: str
):
    multiple_lines = type(log_text) is list

    with open(path, mode) as file:
        if multiple_lines:
            for line in log_text:
                file.write(f"[{entry_time_stamp}]: {line}\n")
        else:
            file.write(f"[{entry_time_stamp}]: {log_text}\n")


def __write_to_log_file_verbose(
    log_text: str | List[str], entry_time_stamp: str, mode: str, path: str
):
    multiple_lines = type(log_text) is list

    with open(path, mode) as file:
        if multiple_lines:
            for line in log_text:
                file.write(f"[{entry_time_stamp}]: {line}\n")
                print(f"[{entry_time_stamp}]: {line}")
        else:
            file.write(f"[{entry_time_stamp}]: {log_text}\n")
            print(f"[{entry_time_stamp}]: {log_text}")


def log_footer(total_time: float):
    log_file_path = log_paths["runtime"]

    text = [
        "=" * 20,
        "Process executed successfully.",
        f"Total time spent in seconds: {total_time}",
        f"Log file saved to: {log_file_path}",
        f"File name: {Path(log_file_path).stem}",
        "=" * 20,
    ]

    log(text)



from liiatools.datasets.social_work_workforce.sample_data import (
    generate_sample_csww_file,
)
from liiatools.csdatatools.util.stream import consume
from liiatools.csdatatools.util.xml import etree, to_xml


def generate_sample(output: str):
    """
    Export a sample file for testing

    :param output: string containing the desired location and name of sample file
    :return: .xml sample file in desired location
    """

    stream = generate_sample_csww_file()
    builder = etree.TreeBuilder()
    stream = to_xml(stream, builder)
    consume(stream)

    element = builder.close()
    element = etree.tostring(element, encoding="utf-8", pretty_print=True)
    try:
        with open(output, "wb") as FILE:
            FILE.write(element)
    except FileNotFoundError:
        print("The file path provided does not exist")



import logging

log = logging.getLogger(__name__)


def to_category(string, categories):
    """
    Matches a string to a category based on categories given in a config file
    the config file should contain a dictionary for each category for this function to loop through
    return blank if no categories found

    :param string: Some string to convert into a category value
    :param categories: A list of dictionaries containing different category:value pairs
    :return: Either a category value, "error" or blank string
    """
    for code in categories:
        if str(string).lower() == str(code["code"]).lower():
            return code["code"]
        if (
            str(string).lower() == str(code["code"]).lower() + ".0"
        ):  # In case integers are read as floats
            return code["code"]
        elif "name" in code:
            if str(code["name"]).lower() in str(string).lower():
                return code["code"]
            elif not string:
                return ""
        elif not string:
            return ""
    return "error"


def to_integer(value, config):
    """
    Convert any strings that should be integers based on the config into integers

    :param value: Some value to convert to an integer
    :param config: The loaded configuration
    :return: Either an integer value or a blank string
    """
    if config == "integer":
        if isinstance(value, str) and value[-2:] == ".0":
            return int(float(value))
        elif value or value == 0:
            return int(value)
        else:
            return ""
    else:
        return value



"""A set of functions that performs operations which convert input files into desired format with valid data."""

import os
from pathlib import Path
from typing import List, Dict, Final

import lxml.etree as etree
from pandas import DataFrame

import liiatools.datasets.social_work_workforce.SWFtools.util.AppLogs as AppLog
import liiatools.datasets.social_work_workforce.SWFtools.dataprocessing.validation.validator as validator
import liiatools.datasets.social_work_workforce.SWFtools.dataprocessing.converter as converter
from liiatools.datasets.social_work_workforce.SWFtools.util.work_path import (
    la_directories,
    flatfile_folder,
)

COLUMNS: Final[List[str]] = [
    "YearCensus",
    "LEA",
    "AgencyWorker",
    "SWENo",
    "FTE",
    "DoB_year",
    "Age",
    "GenderCurrent",
    "Ethnicity",
    "QualInst",
    "QualLevel",
    "StepUpGrad",
    "OrgRole",
    "RoleStartDate",
    "StartOrigin",
    "RoleEndDate",
    "LeaverDestination",
    "FTE30",
    "Cases30",
    "ContractWeeks",
    "FrontlineGrad",
    "CFKSSstatus",
]

COLUMNS_MERGED_FILE: Final[List[str]] = [
    "YearCensus",
    "LEA",
    "LEAName",
    "AgencyWorker",
    "SWENo",
    "FTE",
    "DoB_year",
    "Age",
    "GenderCurrent",
    "Gender",
    "Ethnicity",
    "Ethnicity_Group",
    "Ethnicity_Compact",
    "QualInst",
    "QualLevel",
    "QualLevelName",
    "StepUpGrad",
    "OrgRole",
    "OrgRoleName",
    "RoleStartDate",
    "StartOrigin",
    "RoleEndDate",
    "LeaverDestination",
    "FTE30",
    "Cases30",
    "ContractWeeks",
    "FrontlineGrad",
    "CFKSSstatus",
]


def get_xml_files_from(directory: os.DirEntry) -> List[str]:
    """
    A function that returns a list of XML files that are directly inside the specified directory.
    :param directory: The directory that contains XML files. This is the LA directory
    :return: A list containing the names of the XML files. Each entry has the format: 'filename.xml'
    """
    directory_files = [
        file
        for file in os.listdir(directory)
        if os.path.isfile(os.path.join(directory, file))
    ]
    xml_files = [file for file in directory_files if file.endswith(".xml")]

    return xml_files


def parse_and_validate(
    la_directory: os.DirEntry, xml_file: str
) -> List[Dict[str, str]] | None:
    """
    Parses an XML file, validates it and adds **'LEA'** and **'YearCensus'** common fields.
    :param la_directory: The directory that contains the XML file.
    :param xml_file: The name of the XML file. Format: 'filename.xml'
    :return: Returns a list of dictionaries containing valid worker data. If the file failed validation it returns None
    """
    # === PARSING DATA === #
    AppLog.log(f"Parsing '{xml_file}'...")
    parsed_xml = etree.parse(os.path.join(la_directory, xml_file))

    # === VALIDATING DATA === #
    validation_errors = []

    AppLog.log(f"Validating '{xml_file}'...", console_output=True)
    AppLog.log(f"Validating '{xml_file}'", la_directory.name)

    # Validation failed, continue with next file
    if not validator.is_acceptable(parsed_xml, validation_errors):
        message = f"The file '{xml_file}' failed validation:"
        AppLog.log(message, console_output=True)
        AppLog.log(message, la_directory.name)
        AppLog.log(validation_errors, la_directory.name)
        return None

    AppLog.log(
        f"The file '{xml_file}' passed minimum validation requirement",
        console_output=True,
    )
    AppLog.log(f"Validating worker data...", console_output=True)

    # May contain 'None' values where worker data failed validation
    workers = [
        validator.get_valid_worker_data(worker, validation_errors)
        for worker in parsed_xml.iter("CSWWWorker")
    ]

    # Validation summary data
    total_workers = len(workers)
    valid_count = sum(1 for worker in workers if worker is not None)
    validation_error_count = len(validation_errors)

    # Filter out the 'None' entries
    workers = [worker for worker in workers if worker is not None]

    # Print validation errors if there are any (worker validation)
    if validation_error_count != 0:
        AppLog.log("Worker validation errors: ", la_directory.name)
        AppLog.log(validation_errors, la_directory.name)

    # Print validation summary to LA and application log (runtime)
    messages = [
        f"{valid_count} out of {total_workers} worker records passed validation",
        f"{validation_error_count} worker validation errors.",
    ]
    AppLog.log(messages, la_directory.name)
    AppLog.log(messages, console_output=True)

    # === ADDING EXTENDED FIELDS === #
    # Pre-computed fields for conversion process
    lea = parsed_xml.xpath("//LEA")[0].text
    year_census = parsed_xml.xpath("//Year")[0].text

    converter.column_transfer(workers, lea, year_census)

    return workers


def process_all_input_files():
    """
    Validates XML files, processes worker data, and merges all validated XMLs.
    :return: None
    """
    AppLog.log_section_header(
        "Processing all XML files inside CIN folder", console_output=True
    )

    processed_files_count = 0
    path_merged_file = os.path.join(flatfile_folder, "merged_LA_files.csv")
    path_modified_merged_file = os.path.join(flatfile_folder, "merged_modified.csv")

    for la_directory in la_directories:
        AppLog.log(f"Processing XML files inside: {la_directory.path}")

        xml_files = get_xml_files_from(la_directory)

        for xml_file in xml_files:
            workers = parse_and_validate(la_directory, xml_file)

            if workers is None:
                continue

            # === PROCESSING WORKER DATA === #
            AppLog.log("Processing worker data...", console_output=True)
            AppLog.log("Processing worker data...", la_directory.name)
            for worker in workers:
                converter.swe_hash(worker)
                converter.convert_dates(worker)

            # === WRITING TO CSV === #
            path_file = os.path.join(
                flatfile_folder, la_directory.name, Path(xml_file).stem + ".csv"
            )
            data_frame = DataFrame(workers, columns=COLUMNS_MERGED_FILE)

            # Write single file
            data_frame.to_csv(path_file, index=False, columns=COLUMNS)

            # Write to common file (merged records file)
            # Either append data to existing file or write to file if it's the first file to be processed
            if processed_files_count != 0:
                # header = False (do not append column names again)
                # columns = COLUMNS | COLUMNS_MERGED_FILE (what columns/fields to write)
                # mode is the same as Python file modes ('w' write is default)
                data_frame.to_csv(
                    path_merged_file,
                    index=False,
                    header=False,
                    columns=COLUMNS,
                    mode="a",
                )
                data_frame.to_csv(
                    path_modified_merged_file,
                    index=False,
                    header=False,
                    columns=COLUMNS_MERGED_FILE,
                    mode="a",
                )
            else:
                data_frame.to_csv(path_merged_file, index=False, columns=COLUMNS)
                data_frame.to_csv(
                    path_modified_merged_file, index=False, columns=COLUMNS_MERGED_FILE
                )

            processed_files_count = processed_files_count + 1

    AppLog.log(
        f"Finished processing {processed_files_count} files in {len(la_directories)} directories.",
        console_output=True,
    )



import click as click
from pathlib import Path
import yaml
import logging
import click_log

from sfdata_stream_parser.parser import openpyxl
from liiatools.datasets.annex_a.lds_annexa_clean import (
    configuration as clean_config,
    cleaner,
    degrade,
    logger,
    populate,
    file_creator,
)
from liiatools.datasets.annex_a.lds_annexa_la_agg import configuration as agg_config
from liiatools.datasets.annex_a.lds_annexa_la_agg import process as agg_process

from liiatools.datasets.annex_a.lds_annexa_pan_agg import configuration as pan_config
from liiatools.datasets.annex_a.lds_annexa_pan_agg import process as pan_process

from liiatools.spec import common as common_asset_dir
from liiatools.datasets.shared_functions.common import (
    flip_dict,
    check_file_type,
    supported_file_types,
)
from sfdata_stream_parser.filters.column_headers import promote_first_row

log = logging.getLogger()
click_log.basic_config(log)

COMMON_CONFIG_DIR = Path(common_asset_dir.__file__).parent
# Get all the possible LA codes that could be used
with open(f"{COMMON_CONFIG_DIR}/LA-codes.yml") as las:
    la_list = list(yaml.full_load(las)["data_codes"].values())


@click.group()
def annex_a():
    """Functions for cleaning, minimising and aggregating Annex A files"""
    pass


@annex_a.command()
@click.option(
    "--i",
    "input",
    required=True,
    type=str,
    help="A string specifying the input file location, including the file name and suffix, usable by a pathlib Path function",
)
@click.option(
    "--la_code",
    required=True,
    type=click.Choice(la_list, case_sensitive=False),
    help="A three letter code, specifying the local authority that deposited the file",
)
@click.option(
    "--la_log_dir",
    required=True,
    type=str,
    help="A string specifying the location that the log files for the LA should be output, usable by a pathlib Path function.",
)
@click.option(
    "--o",
    "output",
    required=True,
    type=str,
    help="A string specifying the output directory location",
)
@click_log.simple_verbosity_option(log)
def cleanfile(input, la_code, la_log_dir, output):
    """
    Cleans input Annex A xlsx files according to config and outputs cleaned xlsx files.
    :param input: should specify the input file location, including file name and suffix, and be usable by a Path function
    :param la_code: should be a three-letter string for the local authority depositing the file
    :param la_log_dir: should specify the path to the local authority's log folder
    :param output: should specify the path to the output folder
    :return: None
    """

    # Configuration
    filename = Path(input).resolve().stem
    config = clean_config.Config()
    la_name = flip_dict(config["data_codes"])[la_code]
    if (
        check_file_type(
            input,
            file_types=[".xlsx", ".xlsm"],
            supported_file_types=supported_file_types,
            la_log_dir=la_log_dir,
        )
        == "incorrect file type"
    ):
        return

    # Open & Parse file
    stream = openpyxl.parse_sheets(input)
    stream = [ev.from_event(ev, la_code=la_code, filename=filename) for ev in stream]
    stream = promote_first_row(stream)

    # Configure Stream
    stream = clean_config.configure_stream(stream, config)

    # Clean stream
    stream = cleaner.clean(stream)
    stream = degrade.degrade(stream)
    stream = logger.log_errors(stream)
    stream = populate.create_la_child_id(stream, la_code=la_code)

    # Output result
    stream = file_creator.save_stream(stream, la_name, output)
    stream = logger.save_errors_la(stream, la_log_dir=la_log_dir)
    list(stream)


@annex_a.command()
@click.option(
    "--i",
    "input",
    required=True,
    type=str,
    help="A string specifying the input file location, including the file name and suffix, usable by a pathlib Path function",
)
@click.option(
    "--o",
    "output",
    required=True,
    type=str,
    help="A string specifying the output directory location",
)
def la_agg(input, output):
    """
    Joins data from newly cleaned Annex A file (output of cleanfile()) to existing Annex A data for the depositing local authority
    :param input: a string specifying the input file location of the newly cleaned file, including file name and suffix, usable by a Path function
    :param output: a string specifying the local authority's Annex A output directory, usable by a Path function
    :return: None
    """

    # Configuration
    config = agg_config.Config()

    # Open cleaned file as dictionary
    aa_dict = agg_process.split_file(input)

    # Merge with existing LA file, if any
    list_columns = config["sort_order"]
    aa_dict = agg_process.sort_dict(aa_dict, list_columns=list_columns)
    aa_dict = agg_process.merge_la_files(output, aa_dict)

    # Remove duplicate data and data older than retention period
    dedup = config["dedup"]
    dates = config["dates"]
    index_date = config["index_date"]
    aa_dict = agg_process.deduplicate(aa_dict, dedup=dedup)
    aa_dict = agg_process.convert_datetimes(aa_dict, dates=dates)
    aa_dict = agg_process.remove_old_data(aa_dict, index_date=index_date)
    aa_dict = agg_process.convert_dates(aa_dict, dates=dates)

    # Output result
    agg_process.export_file(output, aa_dict)


@annex_a.command()
@click.option(
    "--i",
    "input",
    required=True,
    type=str,
    help="A string specifying the input file location, including the file name and suffix, usable by a pathlib Path function",
)
@click.option(
    "--la_code",
    required=True,
    type=click.Choice(la_list, case_sensitive=False),
    help="A three letter code, specifying the local authority that deposited the file",
)
@click.option(
    "--o",
    "output",
    required=True,
    type=str,
    help="A string specifying the output directory location",
)
def pan_agg(input, la_code, output):
    """
    Merges data from newly merged Annex A file (output of la_agg()) to existing pan-London Annex A data
    :param input: a string specifying the input file location of the newly merged file, including file name and suffix, usable by a Path function
    :param la_code: should be a three-letter string for the local authority whose data is to be merged
    :param output: a string specifying the pan-London Annex A output directory, usable by a Path function
    :return: None
    """

    # Configuration
    config = pan_config.Config()
    la_name = flip_dict(config["data_codes"])[la_code]

    # Open merged file as dictionary
    pan_dict = pan_process.split_file(input)

    # Minimise data in merged file following schema
    minimise = config["minimise"]
    pan_dict = pan_process.data_minimisation(pan_dict, minimise=minimise)

    # Merge with existing pan-London file, if any
    pan_dict = pan_process.merge_agg_files(output, pan_dict, la_name)

    # Format and write file to folder
    dates = config["dates"]
    pan_dict = pan_process.convert_dates(pan_dict, dates=dates)
    pan_process.export_file(output, pan_dict)



from pathlib import Path
import logging
import datetime
import os
import yaml
from string import Template

from liiatools.spec import cin_census as cin_asset_dir
from liiatools.spec import common as common_asset_dir

log = logging.getLogger(__name__)

DEFAULT_CONFIG_DIR = Path(cin_asset_dir.__file__).parent
COMMON_CONFIG_DIR = Path(common_asset_dir.__file__).parent


class Config(dict):
    def __init__(self, config_files=None):
        super().__init__()

        if not config_files:
            config_files = [
                "PAN_AGG_CONFIG",
                "LA_CODES_CONFIG",
            ]

        for file in config_files:
            if file == "PAN_AGG_CONFIG":
                file = DEFAULT_CONFIG_DIR / "agg.yml"
            elif file == "LA_CODES_CONFIG":
                file = COMMON_CONFIG_DIR / "LA-codes.yml"
            self.load_config(file, conditional=False)

        self["config_date"] = datetime.datetime.now().isoformat()
        try:
            self["username"] = os.getlogin()
        except OSError:
            # This happens when tests are not run under a login shell, e.g. CI pipeline
            pass

    def load_config(self, filename, conditional=False, warn=False):
        """
        Load configuration from yaml file. Any loaded configuration
        is only set if the values don't already exist in CONFIG.

        Files can contain ${} placeholders following the Python string.Template format.
        The context will include any keys already existing in the configuration, any keys
        from the current file - however, if these include placeholders, the placeholders
        will not be replaced. Finally, environment variables can be referenced with
        `os_environ_VARIABLE_NAME`.

        Keyword arguments:
        filename -- Filename to load from
        conditional -- If True, ignore file if it doesn't exist. If False, fail. (default False)
        """
        if conditional and not os.path.isfile(filename):
            if warn:
                log.warning("Missing optional file {}".format(filename))

            return

        with open(filename) as FILE:
            user_config = yaml.load(FILE, Loader=yaml.FullLoader)

        log.info(
            "Loading {} configuration values from '{}'.".format(
                len(user_config), filename
            )
        )

        environment_dict = {"os_environ_{}".format(k): v for k, v in os.environ.items()}

        variables = dict(self)
        variables.update(user_config)
        variables.update(environment_dict)

        with open(filename, "rt") as FILE:
            user_config_string = FILE.read()

        user_config_template = Template(user_config_string)
        user_config_string = user_config_template.substitute(variables)

        user_config = yaml.load(user_config_string, Loader=yaml.FullLoader)

        self.update(user_config)



from functools import cached_property
from pathlib import Path

from xmlschema import XMLSchema

from liiatools.spec import social_work_workforce as social_work_workforce_dir


class Schema:
    def __init__(self, year: int = 2022):
        self.__year = year

    @cached_property
    def schema(self) -> XMLSchema:
        return XMLSchema(
            Path(social_work_workforce_dir.__file__).parent
            / f"social_work_workforce_{self.__year}.xsd"
        )



import logging
import datetime
import os
from pathlib import Path
import yaml
from string import Template

from sfdata_stream_parser import events
from sfdata_stream_parser.filters.generic import streamfilter, pass_event
from sfdata_stream_parser.checks import type_check

from liiatools.datasets.s903.lds_ssda903_clean.columns import column_names
from liiatools.datasets.shared_functions.common import inherit_property
from liiatools.spec import s903 as s903_asset_dir
from liiatools.spec import common as common_asset_dir

log = logging.getLogger(__name__)

DEFAULT_CONFIG_DIR = Path(s903_asset_dir.__file__).parent
SHARED_CONFIG_DIR = Path(common_asset_dir.__file__).parent


@streamfilter(check=type_check(events.StartTable), fail_function=pass_event)
def add_table_name(event):
    """
    Match the loaded table name against one of the 10 903 file names

    :param event: A filtered list of event objects of type StartTable
    :return: An updated list of event objects
    """
    for table_name, expected_columns in column_names.items():
        if set(expected_columns).issubset(set(event.headers)):
            return event.from_event(
                event, expected_columns=expected_columns, table_name=table_name
            )
    return event


@streamfilter(check=type_check(events.Cell), fail_function=pass_event)
def match_config_to_cell(event, config):
    """
    Match the cell to the config file given the table name and cell header
    the config file should be a set of dictionaries for each table, headers within those tables
    and config rules for those headers

    :param event: A filtered list of event objects of type Cell
    :param config: The loaded configuration to use
    :return: An updated list of event objects
    """
    try:
        table_config = config[event.table_name]
        config_dict = table_config[event.header]
        return event.from_event(event, config_dict=config_dict)
    except (
        AttributeError,
        KeyError,
        TypeError,
    ):  # Raise in case there is no config item for the given table name and cell header
        return event


def configure_stream(stream, config):
    """
    Loading and matching the configuration with the loaded stream

    :param stream: Set of events to parse
    :param config: The loaded configuration
    :return: An updated set of events/stream with matched configuration
    """
    stream = add_table_name(stream)
    stream = inherit_property(stream, "table_name")
    stream = inherit_property(stream, "expected_columns")
    stream = match_config_to_cell(stream, config=config["column_map"])
    return stream


class Config(dict):
    def __init__(self, year, *config_files):
        super().__init__()
        if not config_files:
            config_files = ["DEFAULT_COLUMN_MAP", "DEFAULT_LA_MAP"]

        for file in config_files:
            if file == "DEFAULT_COLUMN_MAP":
                file = DEFAULT_CONFIG_DIR / f"SSDA903_schema_{year}.yml"
            elif file == "DEFAULT_LA_MAP":
                file = SHARED_CONFIG_DIR / "LA-codes.yml"
            self.load_config(file, conditional=False)

        self["config_date"] = datetime.datetime.now().isoformat()
        try:
            self["username"] = os.getlogin()
        except OSError:
            # This happens when tests are not run under a login shell, e.g. CI pipeline
            pass

    def load_config(self, filename, conditional=False, warn=False):
        """
        Load configuration from yaml file. Any loaded configuration
        is only set if the values don't already exist in CONFIG.

        Files can contain ${} placeholders following the Python string.Template format.
        The context will include any keys already existing in the configuration, any keys
        from the current file - however, if these include placeholders, the placeholders
        will not be replaced. Finally, environment variables can be referenced with
        `os_environ_VARIABLE_NAME`.

        Keyword arguments:
        filename -- Filename to load from
        conditional -- If True, ignore file if it doesn't exist. If False, fail. (default False)
        """
        if conditional and not os.path.isfile(filename):
            if warn:
                log.warning("Missing optional file {}".format(filename))

            return

        with open(filename) as FILE:
            user_config = yaml.load(FILE, Loader=yaml.FullLoader)

        log.info(
            "Loading {} configuration values from '{}'.".format(
                len(user_config), filename
            )
        )

        environment_dict = {"os_environ_{}".format(k): v for k, v in os.environ.items()}

        variables = dict(self)
        variables.update(user_config)
        variables.update(environment_dict)

        with open(filename, "rt") as FILE:
            user_config_string = FILE.read()

        user_config_template = Template(user_config_string)
        user_config_string = user_config_template.substitute(variables)

        user_config = yaml.load(user_config_string, Loader=yaml.FullLoader)

        self.update(user_config)



import logging
from datetime import datetime
from pathlib import Path

from sfdata_stream_parser.filters.generic import streamfilter, events

log = logging.getLogger(__name__)


def inherit_LAchildID(stream):
    """
    Apply the LAchildID to all elements within <Child></Child>

    :param stream: A filtered list of event objects
    :return: An updated list of event objects
    """
    child_id = child_events = None
    for event in stream:
        if isinstance(event, events.StartElement) and event.tag == "Child":
            child_events = [event]
        elif isinstance(event, events.EndElement) and event.tag == "Child":
            child_events.append(event)
            yield from [e.from_event(e, LAchildID=child_id) for e in child_events]
            child_id = child_events = None
        elif isinstance(event, events.TextNode):
            try:
                child_events.append(event)
                if event.schema.name == "LAchildID":
                    child_id = event.text
            except AttributeError:  # Raised in case there is no LAchildID or event is None
                pass
        elif child_events:
            child_events.append(event)
        else:
            yield event


@streamfilter(check=lambda x: True)
def counter(event, counter_check, value_error, structural_error, blank_error):
    """
    Count the invalid simple nodes storing their name and LAchildID data

    :param event: A filtered list of event objects
    :param counter_check: A function to identify which events to check
    :param value_error: An empty list to store the invalid element information
    :param structural_error: An empty list to store the invalid structure information
    :param blank_error: An empty list to store the blank element information
    :return: The same filtered list of event objects
    """
    if counter_check(event) and len(event.node) == 0:
        if (
            getattr(event, "LAchildID", None) is not None
        ):  # In case there are errors in the <Header> node as none
            # of these will have an LAchildID assigned
            if hasattr(event, "validation_message"):
                blank_error.append(
                    f"LAchildID: {event.LAchildID}, Node: {event.schema.name}"
                )
            elif hasattr(event.schema, "name"):
                value_error.append(
                    f"LAchildID: {event.LAchildID}, Node: {event.schema.name}"
                )
            else:
                structural_error.append(
                    f"LAchildID: {event.LAchildID}, Node: {event.tag}"
                )
        else:
            if hasattr(event, "validation_message"):
                blank_error.append(
                    f"LAchildID: blank, Node: {event.schema.name}"
                )
            elif hasattr(event.schema, "name"):
                value_error.append(f"LAchildID: blank, Node: {event.schema.name}")
            else:
                structural_error.append(f"LAchildID: blank, Node: {event.tag}")
    return event


def save_errors_la(
    input,
    value_error,
    structural_error,
    blank_error,
    LAchildID_error,
    field_error,
    la_log_dir,
):
    """
    Save the errors as a text file in the Local Authority Logs directory
    only save the errors if there is at least one

    :param input: The input file location, including file name and suffix, and be usable by a Path function
    :param value_error: A list of invalid element information
    :param structural_error: A list of invalid structure information
    :param field_error: A list of missing fields
    :param blank_error: A list of blank fields that should have contained data
    :param LAchildID_error: A list of missing LAchildID information
    :param la_log_dir: Location to save the gathered error logs
    :return: An updated list of event objects
    """
    filename = str(Path(input).resolve().stem)
    start_time = f"{datetime.now():%Y-%m-%dT%H%M%SZ}"
    if (
        value_error
        or structural_error
        or field_error
        or blank_error
        or LAchildID_error
    ):
        with open(
            f"{Path(la_log_dir, filename)}_error_log_{start_time}.txt",
            "a",
        ) as f:
            f.write(filename)
            f.write("\n")
            if value_error:
                f.write(
                    "Node information for data that has been removed because it was found to be invalid"
                )
                f.write("\n")
                for item in value_error:
                    f.write(item)
                    f.write("\n")
                f.write("\n")
            if structural_error:
                f.write(
                    "Node information for fields that has been removed because they did not match the "
                    "expected structure"
                )
                f.write("\n")
                for item in structural_error:
                    f.write(item)
                    f.write("\n")
                f.write("\n")
            if blank_error:
                f.write(
                    "Node information for blank fields that should have contained data"
                )
                f.write("\n")
                for item in blank_error:
                    f.write(item)
                    f.write("\n")
                f.write("\n")
            if LAchildID_error:
                LAchildID_error = list(
                    dict.fromkeys(LAchildID_error)
                )  # Remove duplicate information from list but
                # keep order
                for item in LAchildID_error:
                    f.write(item)
                    f.write("\n")
            if field_error:
                field_error = list(
                    dict.fromkeys(field_error)
                )  # Remove duplicate information from list but keep order
                for item in field_error:
                    f.write(item)
                    f.write("\n")
                f.write("\n")
